# -*- coding: utf-8 -*-
"""
The :class:`Control` object is the central part of pyMolDyn.
It is responsible for initiating the calculation process and visualizing
the results.
It contains instances of :class:`core.calculation.calculation.Calculation` and
:class:`visualization.visualization.Visualization` and manages the
interaction between them.

The calculation can be started in an different thread. To do that,
the ``calculationcallback`` attribute can be set
to a custom function. This custom function has to take two parameters:
``func`` and ``settings``, and has to call ``func`` with the parameter
``settings``. Here is an easy example::

    def callback(func, settings):
        func(settings)

    control.calculationcallback = callback

And with threads::

    class CalculationThread(QtCore.QThread):
        def __init__(self, parent, func, settings):
            QtCore.QThread.__init__(self, parent)
            self.func = func
            self.settings = settings

        def run(self):
            self.func(self.settings)

    class GUI:
        def __init__(self, ...):
            ...
            self.control.calculationcallback = self.calculationcallback

        def calculationcallback(self, func, settings):
            thread = CalculationThread(self, func, settings)
            thread.finished.connect(self.control.update)
            thread.start()
"""


import os.path
import threading
from ..config.configuration import config
from ..core import calculation as calculation
from .. import visualization


# TODO: make an instance of this available everywhere


class Control(object):
    """
    The central controller class that contains the application logic.
    It contains the following attributes:

        `calculation` : :class:`core.calculation.calculation.Calculation`

        `visualization` : :class:`visualization.visualization.Visualization`
    """

    def __init__(self):
        self.config = config
        if not os.path.isdir(os.path.expanduser(config.Path.cache_dir)):
            os.makedirs(os.path.expanduser(config.Path.cache_dir))
        self._calculation = calculation.Calculation()
        self._visualization = None  # will be initialized when needed
        self.results = None
        self.calculationcallback = Control.defaultcallback
        self.lock = threading.Lock()

    def _calculate(self, settings):
        with self.lock:
            self.results = None
            self.results = self.calculation.calculate(settings)

    def calculate(self, settings):
        """
        Start a calculation and create prepare the results to be visualized.
        To make it possible to start the calculation in a thread, the
        function ``self.calculation`` is called, which then starts the
        calculation itself.

        **Parameters:**
            `settings` :
                :class:`core.calculation.calculation.CalculationSettings` object
        """
        # TODO: only call when something is calculated
        self.calculationcallback(self._calculate, settings.copy())

    def update(self, was_successful=lambda: True):
        """
        Visualize previously calculated results. It has to be called from
        the same thread which uses the OpenGL context, usually the
        event handler of the GUI.
        """
        with self.lock:
            if was_successful and self.results is not None:
                self.visualization.setresults(self.results[-1][-1])

    def visualize(self, filename, frame, resolution=None):
        """
        Visualize the given frame. if the ``resolution`` parameter is not set,
        this function uses the highest resolution available to show
        calculation results.
        """
        results = self.calculation.getresults(filename, frame, resolution)
        self.results = [[results]]
        self.update()

    @staticmethod
    def defaultcallback(func, settings):
        func(settings)

    @property
    def calculation(self):
        return self._calculation

    @property
    def visualization(self):
        if self._visualization is None:
            self._visualization = visualization.Visualization()
        return self._visualization



# -*- coding: utf-8 -*-
"""
This module contains classes that are used to store data in pyMolDyn.
Most of them can be read and written to hdf5 files.
"""

import collections
import numpy as np
import sys
import os
from datetime import datetime
import dateutil.parser
import h5py
from . import volumes
from ..config.configuration import config
from ..util.logger import Logger
from . import elements
from . import bonds

try:
    import openbabel

    USE_PYBEL = True
    PYBEL_MOLECULE_TYPE = openbabel.pybel.Molecule  # TODO openbabel install
except ImportError:
    USE_PYBEL = False
    PYBEL_MOLECULE_TYPE = None


logger = Logger("core.data")
logger.setstream("default", sys.stdout, Logger.WARNING)


__all__ = [
    "Atoms",
    "Domains",
    "Cavities",
    "Results",
    "FileInfo",
    "ResultInfo",
    "CalculatedFrames",
]


def writedataset(h5group, name, data, overwrite=True):
    """
    Write a dataset to a hdf5 file.
    The `overwrite` parameter controls the behaviour when the dataset
    already exists.
    """
    if name in h5group:
        if overwrite:
            del h5group[name]
        else:
            return False
    h5group[name] = data
    return True


class TimestampList(object):
    """
    A `list`-like structure with a fixed length to store :class:`datetime`
    objects.  For each frame it contains a calculation date or `None`.
    """

    def __init__(self, *args):
        """
        Creates an empty :class:`TimestampList` with a given length
        or copies values from a `list`.

        The constructor can be called in two ways:

        - ``TimestampList(num_frames)`` :
            create an empty :class:`TimestampList` with the length `num_frames`

        - ``TimestampList(list)`` :
            copy the values from the given list of strings
        """
        if isinstance(args[0], list):
            arr = args[0]
            self.timestamps = [None] * len(arr)
            for i, s in enumerate(arr):
                if len(s) > 0:
                    self.timestamps[i] = dateutil.parser.parse(s)
        else:
            num_frames = args[0]
            self.timestamps = [None] * num_frames

    @property
    def num_frames(self):
        return len(self.timestamps)

    def __getitem__(self, index):
        if 0 <= index < len(self.timestamps):
            return self.timestamps[index]
        else:
            return None

    def __setitem__(self, index, value):
        if not isinstance(value, datetime):
            raise ValueError("datetime required")
        self.timestamps[index] = value

    def __len__(self):
        return len(self.timestamps)

    def __iter__(self):
        return iter(self.timestamps)

    def hasdata(self):
        """
        Test if the list contains any data.

        **Returns:**
            If any item is not `None`
        """
        return any(map(lambda x: x is not None, self.timestamps))

    def tostrlist(self):
        """
        Converts the :class:`datetime` objects to strings.
        `None` values are converted to empty strings.

        **Returns:**
            A list of strings
        """
        return ["" if x is None else str(x) for x in self.timestamps]

    def prettystrings(self):
        """
        Converts the :class:`datetime` objects to human-readable strings.
        `None` values are converted to "X".

        **Returns:**
            A list of strings
        """

        def fmt(t):
            if t is None:
                return "X"
            else:
                return t.strftime("%d.%m.%Y %H:%M:%S")

        return map(fmt, self.timestamps)


class CalculatedFrames(object):
    """
    Contains information what results are calculated for what frame.
    The information about the three kinds of results are stored in
    these attributes of the type :class:`TimestampList`:

        `domains`

        `surface_cavities`

        `center_cavities`
    """

    def __init__(self, *args):
        """
        Creates an empty object or reads it from a hdf5 file.

        The constructor can be called in two ways:

        - ``CalculatedFrames(num_frames)`` :
            create an empty `CalculatedFrames` which can store information
            about `num_frames` frames

        - ``CalculatedFrames(hdf5group)`` :
            read the data from this hdf5 group
        """
        if isinstance(args[0], h5py.Group):
            h5group = args[0]
            dom_ts = list(h5group["domains"])
            sur_ts = list(h5group["surface_cavities"])
            cen_ts = list(h5group["center_cavities"])
            num_frames = len(dom_ts)
            self.domains = TimestampList(dom_ts)
            self.surface_cavities = TimestampList(sur_ts)
            self.center_cavities = TimestampList(cen_ts)
        else:
            num_frames = args[0]
            self.domains = TimestampList(num_frames)
            self.surface_cavities = TimestampList(num_frames)
            self.center_cavities = TimestampList(num_frames)

    @property
    def num_frames(self):
        return self.domains.num_frames

    def hasdata(self):
        """
        Test if member contains any data.

        **Returns:**

            If the `hasdata` method of any of the :class:`TimestampList`
            attributes returns `True`
        """
        return (
            self.domains.hasdata()
            or self.surface_cavities.hasdata()
            or self.center_cavities.hasdata()
        )

    def tohdf(self, h5group, overwrite=True):
        """
        Write the data to a hdf5 Group.

        **Parameters:**
            `h5group` :
                the hdf5 group in which the data will be written

            `overwrite` :

                specifies if existing data should be overwritten
        """
        if self.hasdata():
            writedataset(h5group, "domains", self.domains.tostrlist(), overwrite)
            writedataset(
                h5group,
                "surface_cavities",
                self.surface_cavities.tostrlist(),
                overwrite,
            )
            writedataset(
                h5group, "center_cavities", self.center_cavities.tostrlist(), overwrite
            )


class FileInfo(object):
    """
    Contains information about input files:

        `num_frames` :
            number of frames which are stored in the file

        `volumestr` :
            representative string of the volume the atoms are in

        `volume` :
            the corresponding `Volume` object itself

    """

    def __init__(self, num_frames=None, volumestr=None, volume_guessed=False):
        """
        **Parameters:**
            `num_frames` :
                number of frames which are stored in the file

            `volumestr` :
                representative string of the volume the atoms are in
        """
        self.num_frames = num_frames
        self.volumestr = volumestr
        self._volume = None
        self.volume_guessed = volume_guessed

    @property
    def volume(self):
        if self._volume is None and self.volumestr is not None:
            self._volume = volumes.Volume.fromstring(self.volumestr)
        return self._volume


class ResultInfo(FileInfo):
    """
    Information about a :class:`core.file.ResultFile`:

        `num_frames` :
            number of frames which are stored in the file

        `volumestr` :
            representative string of the volume the atoms are in

        `volume` :
            the corresponding `Volume` object itself

        `sourcefilepath` :
            path of the file which contains the original input data

        `calculatedframes` :
            dictionary which associates a resolution with a
            `CalculatedFrames` object
    """

    def __init__(self, *args):
        """
        The constructor can be called in two ways:

        - ``ResultInfo()`` :
            create an empty :class:`ResultInfo` object

        - ``ResultInfo(hdf5group)`` :
            read the data from this hdf5 group
        """
        super().__init__()
        self.sourcefilepath = None
        self.calculatedframes = dict()
        if len(args) > 0 and isinstance(args[0], h5py.Group):
            h5group = args[0]
            self.num_frames = int(h5group.attrs["num_frames"])
            self.volumestr = str(h5group.attrs["volume"])
            if "sourcefile" in h5group.attrs:
                self.sourcefilepath = str(h5group.attrs["sourcefile"])
            else:
                self.sourcefilepath = None
            for name, subgroup in h5group.items():
                if not name.startswith("resolution"):
                    continue
                resolution = int(name[10:])
                self.calculatedframes[resolution] = CalculatedFrames(subgroup)

    def __getitem__(self, resolution):
        """
        Get a :class:`CalculatedFrames` object for the specified resulotion.

        **Parameters:**
            `resolution`:
                the resolution for which the information will be queried

        **Returns:**
             A :class:`CalculatedFrames` object with information about
             existing calculations.
        """
        if resolution not in self.calculatedframes:
            self.calculatedframes[resolution] = CalculatedFrames(self.num_frames)
        return self.calculatedframes[resolution]

    def __contains__(self, resolution):
        """
        Check if any information about the given resolution is available.

        **Parameters:**
            `resolution`:
                the resolution for which the information will be queried

        **Returns:**
             If a :class:`CalculatedFrames` object for this resolution exists
        """
        return resolution in self.calculatedframes

    def resolutions(self):
        return self.calculatedframes.keys()

    def tohdf(self, h5group, overwrite=True):
        """
        Write the data to a hdf5 Group.

        **Parameters:**
            `h5group` :
                the hdf5 group in which the data will be written

            `overwrite` :
                specifies if existing data should be overwritten
        """
        h5group.attrs["num_frames"] = self.num_frames
        h5group.attrs["volume"] = self.volumestr
        if self.sourcefilepath is not None:
            h5group.attrs["sourcefile"] = self.sourcefilepath
        elif "sourcefile" in h5group.attrs:
            del h5group.attrs["sourcefile"]
        for resolution, info in self.calculatedframes.items():
            if info.hasdata():
                subgroup = h5group.require_group("resolution{}".format(resolution))
                info.tohdf(subgroup, overwrite)


class Atoms(object):
    """
    This class represents a list of atoms and their properties:

        - `number`

        - `positions`

        - `volume` :
            the volume that contains the atoms

        - `radii` :
            cavity cutoff radius

        - `sorted_positions` :
            `positions` sorted from largest to smallest radius

        - `sorted_radii` :
            unique `radii` sorted from largest to smallest

        - `radii_as_indices` :
            indices to associate an atom in `sorted_positions` with an element
            of `sorted_radii`
    """

    def __init__(self, *args):
        """
        The constructor can be called in three ways:

        - ``Atoms(positions, radii, elements, volume)`` :
            create the object using the given data

        - ``Atoms(molecule, volume)`` :
            read the data from this :class:`pybel.Molecule` object

        - ``Atoms(hdf5group)`` :
            read the data from this hdf5 group
        """
        if isinstance(args[0], h5py.Group):
            h5group = args[0]
            positions = h5group["positions"]
            radii = h5group["radii"]
            if "elements" in h5group:
                elements = h5group["elements"]
            else:
                logger.warn(
                    "Dataset 'elements' not found. Using 'atom' as default value"
                )
                elements = np.empty(len(radii), dtype="|S4")
                elements[:] = "atom"
            if "volume" in h5group.attrs:
                volume = h5group.attrs["volume"]
            else:
                volume = h5group.parent.attrs["volume"]
            volume = volumes.Volume.fromstring(str(volume))
        else:
            # in these two cases atom positions may be outside of the volume
            if USE_PYBEL and isinstance(args[0], PYBEL_MOLECULE_TYPE):
                molecule = args[0]
                if len(args) > 1:
                    volume = args[1]
                else:
                    volume = None
                positions = map(lambda atom: atom.coords, molecule)
                elements = map(lambda atom: elements.symbols[atom.atomicnum], molecule)
                radii = None
            else:
                positions = args[0]
                radii = args[1]
                elements = args[2]
                volume = args[3]

            if isinstance(volume, str):
                volume = volumes.Volume.fromstring(volume)
            if volume is not None:
                positions = map(lambda pos: volume.get_equivalent_point(pos), positions)
        self.volume = volume
        self.positions = np.asarray(list(positions), dtype=np.float64)
        self.number = self.positions.shape[0]
        self.elements = np.asarray(elements, dtype="|S4")
        self.radii = radii

        self._covalence_radii = None
        self._covalence_radii_by_element = None
        self._bonds = None
        self._colors = None

    @property
    def covalence_radii(self):
        if self._covalence_radii is None:
            covalence_radii = np.zeros(self.number, np.float32)
            for i, element in enumerate(self.elements):
                element_number = elements.numbers[element.upper().decode("utf-8")]
                covalence_radii[i] = elements.radii[element_number]
            self._covalence_radii = covalence_radii
        return self._covalence_radii

    @property
    def covalence_radii_by_element(self):
        covalence_radii = self.covalence_radii
        if self._covalence_radii_by_element is None:
            self._covalence_radii_by_element = dict(
                (element, covalence_radius)
                for element, covalence_radius in zip(self.elements, covalence_radii)
            )
        return self._covalence_radii_by_element

    @property
    def bonds(self):
        if self._bonds is None:
            self._bonds = bonds.get_bonds_with_radii(self, 1.15)
            self._bonds = bonds.get_bonds_symetric_indicies(self._bonds)
        return self._bonds

    @property
    def colors(self):
        if self._colors is None:
            colors = np.zeros((self.number, 3), np.float32)
            for i, element in enumerate(self.elements):
                element_number = elements.numbers[element.decode("utf-8").upper()]
                colors[i] = elements.colors[element_number]
            self._colors = colors / 255
        return self._colors

    @property
    def radii(self):
        return self._radii

    @radii.setter
    def radii(self, values):
        if values is None:
            self._radii = (
                np.ones((self.number), dtype=np.float64)
                * config.Computation.std_cutoff_radius
            )
        elif isinstance(values, collections.abc.Mapping):
            radii = [values[elem] for elem in self.elements]
            self._radii = np.asarray(radii, dtype=np.float64)
        elif isinstance(values, collections.abc.Iterable):
            self._radii = np.asarray(values, dtype=np.float64)
        else:
            self._radii = np.ones((self.number), dtype=np.float64) * values
        indices = np.argsort(-self._radii, kind="mergesort")
        self.sorted_positions = self.positions[indices]
        unique_radii, indices = np.unique(-self._radii, return_inverse=True)
        self.sorted_radii = -unique_radii
        self.radii_as_indices = np.sort(indices)

    def tohdf(self, h5group, overwrite=True):
        """
        Write the data to a hdf5 Group.

        **Parameters:**
            `h5group` :
                the hdf5 group in which the data will be written

            `overwrite` :
                specifies if existing data should be overwritten
        """
        h5group.parent.attrs["volume"] = str(self.volume)
        writedataset(h5group, "positions", self.positions, overwrite)
        writedataset(h5group, "radii", self.radii, overwrite)
        if np.any(self.elements == "atom"):
            logger.warn("Atom.elements contains default values. Not writing dataset.")
        else:
            writedataset(h5group, "elements", self.elements, overwrite)

    def totxt(self, fmt):
        bond_file_name = fmt.format(property="bonds")
        bond_angle_file_name = fmt.format(property="bond_angles")
        bond_chain_angle_file_name = fmt.format(property="bond_dihedral_angles")

        bond_angles, bond_chain_angles = bonds.calculate_bond_angles(self, self.bonds)

        with open(bond_file_name, "w") as outfile:
            for source_index, target_indices in enumerate(self.bonds):
                for target_index in target_indices:
                    outfile.write("{} {}\n".format(source_index + 1, target_index + 1))
        with open(bond_angle_file_name, "w") as outfile:
            for bond1, bond2 in bond_angles.keys():
                if bond1[0] > bond2[1]:
                    outfile.write(
                        "{} {} {} {}\n".format(
                            bond1[0] + 1,
                            bond1[1] + 1,
                            bond2[1] + 1,
                            bond_angles[bond1, bond2],
                        )
                    )
        with open(bond_chain_angle_file_name, "w") as outfile:
            for bond_chain, angle in bond_chain_angles.items():
                outfile.write(
                    "{} {} {} {}".format(*[index + 1 for index in bond_chain])
                )
                outfile.write(" {}\n".format(angle))


class CavitiesBase(object):
    """
    Base class to store multiple surface-based objects in the 3-dimensional
    space. The :class:`Domains` and :class:`Cavities` class inherit from it.
    """

    def __init__(self, *args):
        """
        The constructor can be called in two ways:

        - ``CavitiesBase(timestamp, volumes, surface_areas, triangles)`` :
            create the object using the given data

        - ``CavitiesBase(hdf5group)`` :
            read the data from this hdf5 group
        """
        if isinstance(args[0], h5py.Group):

            def getobj_from_h5group(attr):
                if attr in h5group:
                    return h5group[attr]
                else:
                    return None

            h5group = args[0]
            timestamp = dateutil.parser.parse(h5group.attrs["timestamp"])
            number = int(h5group.attrs["number"])
            volumes = getobj_from_h5group("volumes")
            surface_areas = getobj_from_h5group("surface_areas")
            triangles = [None] * number
            for i in range(number):
                triangles[i] = getobj_from_h5group("triangles{}".format(i))
            mass_centers = getobj_from_h5group("mass_centers")
            squared_gyration_radii = getobj_from_h5group("squared_gyration_radii")
            asphericities = getobj_from_h5group("asphericities")
            acylindricities = getobj_from_h5group("acylindricities")
            anisotropies = getobj_from_h5group("anisotropies")
            characteristic_radii = getobj_from_h5group("characteristic_radii")
            cyclic_area_indices = getobj_from_h5group("cyclic_area_indices")
        else:
            (
                timestamp,
                volumes,
                surface_areas,
                triangles,
                mass_centers,
                squared_gyration_radii,
                asphericities,
                acylindricities,
                anisotropies,
                characteristic_radii,
                cyclic_area_indices,
            ) = args[:11]

        if not isinstance(timestamp, datetime):
            timestamp = dateutil.parser.parse(str(timestamp))
        self.timestamp = timestamp
        self.volumes = np.asarray(volumes, dtype=np.float64)
        self.number = len(volumes)
        self.surface_areas = np.asarray(surface_areas, dtype=np.float64)
        self.triangles = [
            np.asarray(triangle, dtype=np.float64) for triangle in triangles
        ]
        self.mass_centers = np.asarray(mass_centers, dtype=np.float64)
        self.squared_gyration_radii = np.asarray(
            squared_gyration_radii, dtype=np.float64
        )
        self.asphericities = np.asarray(asphericities, dtype=np.float64)
        self.acylindricities = np.asarray(acylindricities, dtype=np.float64)
        self.anisotropies = np.asarray(anisotropies, dtype=np.float64)
        self.characteristic_radii = np.asarray(characteristic_radii, dtype=np.float64)
        self.cyclic_area_indices = (
            np.asarray(cyclic_area_indices, dtype=np.int32)
            if cyclic_area_indices is not None
            else np.array([])
        )

    def tohdf(self, h5group, overwrite=True):
        """
        Write the data to a hdf5 Group.

        **Parameters:**
            `h5group` :
                the hdf5 group in which the data will be written

            `overwrite` :
                specifies if existing data should be overwritten
        """
        h5group.attrs["timestamp"] = str(self.timestamp)
        h5group.attrs["number"] = self.number
        writedataset(h5group, "volumes", self.volumes, overwrite)
        writedataset(h5group, "surface_areas", self.surface_areas, overwrite)
        for index, triangles in enumerate(self.triangles):
            writedataset(
                h5group, "triangles{}".format(index), np.array(triangles), overwrite
            )
        writedataset(h5group, "mass_centers", self.mass_centers, overwrite)
        writedataset(
            h5group, "squared_gyration_radii", self.squared_gyration_radii, overwrite
        )
        writedataset(h5group, "asphericities", self.asphericities, overwrite)
        writedataset(h5group, "acylindricities", self.acylindricities, overwrite)
        writedataset(h5group, "anisotropies", self.anisotropies, overwrite)
        writedataset(
            h5group, "characteristic_radii", self.characteristic_radii, overwrite
        )
        writedataset(
            h5group, "cyclic_area_indices", self.cyclic_area_indices, overwrite
        )

    def _export_gyration_parameters(self, fmt):
        mass_centers_file_name = fmt.format(property="centers_of_mass")
        squared_gyration_radii_file_name = fmt.format(property="squared_gyration_radii")
        asphericities_file_name = fmt.format(property="asphericities")
        acylindricities_file_name = fmt.format(property="acylindricities")
        anisotropies_file_name = fmt.format(property="anisotropies")
        characteristic_radii_file_name = fmt.format(property="characteristic_radii")

        mass_centers = self.getattr_normalized("mass_centers")
        squared_gyration_radii = self.getattr_normalized("squared_gyration_radii")
        asphericities = self.getattr_normalized("asphericities")
        acylindricities = self.getattr_normalized("acylindricities")
        anisotropies = self.getattr_normalized("anisotropies")
        characteristic_radii = self.getattr_normalized("characteristic_radii")

        export_filenames = []

        if mass_centers is not None:
            with open(mass_centers_file_name, "w") as outfile:
                for index, mass_center in enumerate(mass_centers, start=1):
                    outfile.write("{} {}\n".format(index, mass_center))
            export_filenames.append(mass_centers_file_name)
        if squared_gyration_radii is not None:
            with open(squared_gyration_radii_file_name, "w") as outfile:
                for index, squared_gyration_radius in enumerate(
                    squared_gyration_radii, start=1
                ):
                    outfile.write("{} {}\n".format(index, squared_gyration_radius))
            export_filenames.append(squared_gyration_radii_file_name)
        if asphericities is not None:
            with open(asphericities_file_name, "w") as outfile:
                for index, asphericity in enumerate(asphericities, start=1):
                    outfile.write("{} {}\n".format(index, asphericity))
            export_filenames.append(asphericities_file_name)
        if acylindricities is not None:
            with open(acylindricities_file_name, "w") as outfile:
                for index, acylindricity in enumerate(acylindricities, start=1):
                    outfile.write("{} {}\n".format(index, acylindricity))
            export_filenames.append(acylindricities_file_name)
        if anisotropies is not None:
            with open(anisotropies_file_name, "w") as outfile:
                for index, anisotropy in enumerate(anisotropies, start=1):
                    outfile.write("{} {}\n".format(index, anisotropy))
            export_filenames.append(anisotropies_file_name)
        if characteristic_radii is not None:
            with open(characteristic_radii_file_name, "w") as outfile:
                for index, characteristic_radius in enumerate(
                    characteristic_radii, start=1
                ):
                    outfile.write("{} {}\n".format(index, characteristic_radius))
            export_filenames.append(characteristic_radii_file_name)

        return export_filenames

    def getattr_normalized(self, attr):
        if hasattr(self, attr):
            value = getattr(self, attr)
            is_numpy_array = isinstance(value, np.ndarray)
            if (is_numpy_array and len(value.shape) > 0) or (
                not is_numpy_array and len(value) > 0
            ):
                return value
        return None


class Domains(CavitiesBase):
    """
    Stores the calculated data about the domains.
    """

    def __init__(self, *args):
        """
        The constructor can be called in three ways:

        - ``Domains(timestamp, volumes, surface_areas, triangles, centers)`` :
            create the object using the given data

        - ``Domains(domaincalculation)`` :
            copy the data from this
            :class:`core.calculation.algorithm.DomainCalculation` object

        - ``Domains(hdf5group)`` :
            read the data from this hdf5 group
        """
        # Import this here to avoid cyclic imports
        from .calculation import algorithm

        if isinstance(args[0], h5py.Group):
            super().__init__(*args)
            h5group = args[0]
            centers = h5group["centers"]
        elif isinstance(args[0], algorithm.DomainCalculation):
            calculation = args[0]
            timestamp = datetime.now()
            volumes = calculation.domain_volumes
            surface_areas = calculation.domain_surface_areas
            triangles = calculation.domain_triangles
            centers = calculation.centers
            discretization = calculation.discretization
            mass_centers = calculation.mass_centers
            squared_gyration_radii = calculation.squared_gyration_radii
            asphericities = calculation.asphericities
            acylindricities = calculation.acylindricities
            anisotropies = calculation.anisotropies
            characteristic_radii = calculation.characteristic_radii
            cyclic_area_indices = calculation.cyclic_area_indices
            super().__init__(
                timestamp,
                volumes,
                surface_areas,
                triangles,
                mass_centers,
                squared_gyration_radii,
                asphericities,
                acylindricities,
                anisotropies,
                characteristic_radii,
                cyclic_area_indices,
            )
        else:
            super().__init__(*args)
            centers = args[4]

        self.centers = np.asarray(centers, dtype=np.int32)
        if "discretization" in locals():
            # TODO: get discretization also from other constructor calls!
            self.continuous_centers = np.array(
                [discretization.discrete_to_continuous(center) for center in centers],
                dtype=np.float64,
            )

    def tohdf(self, h5group, overwrite=True):
        """
        Write the data to a hdf5 Group.

        **Parameters:**
            `h5group` :
                the hdf5 group in which the data will be written

            `overwrite` :
                specifies if existing data should be overwritten
        """
        super(Domains, self).tohdf(h5group, overwrite)
        writedataset(h5group, "centers", self.centers, overwrite)

    def totxt(self, fmt):
        domain_center_file_name = fmt.format(property="centers")
        domain_surface_file_name = fmt.format(property="surface_areas")
        domain_volume_file_name = fmt.format(property="volumes")
        domain_surface_to_volume_ratio_file_name = fmt.format(
            property="surface_area_to_volume_ratios"
        )

        export_filenames = [
            domain_center_file_name,
            domain_surface_file_name,
            domain_volume_file_name,
            domain_surface_to_volume_ratio_file_name,
        ]

        with open(domain_surface_file_name, "w") as outfile:
            for index, surface_area in enumerate(self.surface_areas, start=1):
                outfile.write("{} {}\n".format(index, surface_area))
        with open(domain_volume_file_name, "w") as outfile:
            for index, volume in enumerate(self.volumes, start=1):
                outfile.write("{} {}\n".format(index, volume))
        with open(domain_surface_to_volume_ratio_file_name, "w") as outfile:
            for index, t in enumerate(zip(self.volumes, self.surface_areas), start=1):
                volume, surface_area = t
                outfile.write("{} {}\n".format(index, surface_area / volume))
        continuous_centers = self.getattr_normalized("continuous_centers")
        if continuous_centers is not None:
            with open(domain_center_file_name, "w") as outfile:
                for index, continuous_center in enumerate(continuous_centers, start=1):
                    outfile.write(
                        "{} {} {} {}\n".format(
                            index,
                            continuous_center[0],
                            continuous_center[1],
                            continuous_center[2],
                        )
                    )
        else:
            raise ValueError(
                "No discretization present -> can only access discrete domain centers, no conversion to continuous space possible."
            )

        export_filenames.extend(self._export_gyration_parameters(fmt))

        return export_filenames


class Cavities(CavitiesBase):
    """
    Stores the calculated data about the cavities.
    """

    def __init__(self, *args):
        """
        The constructor can be called in three ways:

        - ``Cavities(timestamp, volumes, surface_areas, triangles, multicavities)`` :
            create the object using the given data

        - ``Cavities(cavitycalculation)`` :
            copy the data from this
            :class:`core.calculation.algorithm.CavityCalculation` object

        - ``Cavities(hdf5group)`` :
            read the data from this hdf5 group
        """
        # Import this here to avoid cyclic imports
        from .calculation import algorithm

        if isinstance(args[0], h5py.Group):
            super().__init__(*args)
            h5group = args[0]
            multicavities = [None] * self.number
            for i in range(self.number):
                multicavities[i] = h5group["multicavities{}".format(i)]
        elif isinstance(args[0], algorithm.CavityCalculation):
            calculation = args[0]
            timestamp = datetime.now()
            volumes = calculation.multicavity_volumes
            surface_areas = calculation.cavity_surface_areas
            triangles = calculation.cavity_triangles
            multicavities = calculation.multicavities
            mass_centers = calculation.mass_centers
            squared_gyration_radii = calculation.squared_gyration_radii
            asphericities = calculation.asphericities
            acylindricities = calculation.acylindricities
            anisotropies = calculation.anisotropies
            characteristic_radii = calculation.characteristic_radii
            cyclic_area_indices = calculation.cyclic_area_indices
            super().__init__(
                timestamp,
                volumes,
                surface_areas,
                triangles,
                mass_centers,
                squared_gyration_radii,
                asphericities,
                acylindricities,
                anisotropies,
                characteristic_radii,
                cyclic_area_indices,
            )
        else:
            super().__init__(*args)
            multicavities = args[4]

        self.multicavities = [None] * self.number
        for index, cavities in enumerate(multicavities):
            # cavities might be a 0-dimensional ndarray of python objects
            if isinstance(cavities, np.ndarray):
                cavities = cavities.tolist()
            self.multicavities[index] = np.array(list(cavities), dtype=np.int32)

    def tohdf(self, h5group, overwrite=True):
        """
        Write the data to a hdf5 Group.

        **Parameters:**
            `h5group` :
                the hdf5 group in which the data will be written

            `overwrite` :
                specifies if existing data should be overwritten
        """
        super(Cavities, self).tohdf(h5group, overwrite)
        for index, cavities in enumerate(self.multicavities):
            writedataset(h5group, "multicavities{}".format(index), cavities, overwrite)

    def totxt(self, fmt):
        cavity_surface_file_name = fmt.format(property="surface_areas")
        cavity_volume_file_name = fmt.format(property="volumes")
        cavity_domains_file_name = fmt.format(property="domain_indices")
        cavity_surface_to_volume_ratio_file_name = fmt.format(
            property="surface_area_to_volume_ratios"
        )

        export_filenames = [
            cavity_surface_file_name,
            cavity_surface_file_name,
            cavity_volume_file_name,
            cavity_surface_to_volume_ratio_file_name,
        ]

        with open(cavity_domains_file_name, "w") as outfile:
            for index, multicavity in enumerate(self.multicavities, start=1):
                outfile.write("{}".format(index))
                for domain_index in multicavity:
                    outfile.write(" {}".format(domain_index + 1))
                outfile.write("\n".format(index))
        with open(cavity_surface_file_name, "w") as outfile:
            for index, surface_area in enumerate(self.surface_areas, start=1):
                outfile.write("{} {}\n".format(index, surface_area))
        with open(cavity_volume_file_name, "w") as outfile:
            for index, volume in enumerate(self.volumes, start=1):
                outfile.write("{} {}\n".format(index, volume))
        with open(cavity_surface_to_volume_ratio_file_name, "w") as outfile:
            for index, t in enumerate(zip(self.volumes, self.surface_areas), start=1):
                volume, surface_area = t
                outfile.write("{} {}\n".format(index, surface_area / volume))

        export_filenames.extend(self._export_gyration_parameters(fmt))

        return export_filenames


class Results(object):
    """
    Container class to store the calculated putput data together with its
    input data. This can be passed to the Visualization:

        - `filepath`
        - `frame`
        - `resolution`
        - `atoms`
        - `domains`
        - `surface_cavities`
        - `center_cavities`
    """

    def __init__(
        self,
        filepath,
        frame,
        resolution,
        atoms,
        domains,
        surface_cavities,
        center_cavities,
    ):
        """
        **Parameters:**
            `filepath` :
                path to the input file

            `frame` :
                frame number

            `resolution`:
                resolution of the discretization

            `atoms` :
                input data

            `domains` :
                the calculated domains or `None`

            `surface_cavities`
                the calculated surface-based cavities or `None`

            `center_cavities`
                the calculated center-based cavities or `None`
        """
        self.filepath = filepath
        self.frame = frame
        self.resolution = resolution
        self.atoms = atoms
        self.domains = domains
        self.surface_cavities = surface_cavities
        self.center_cavities = center_cavities

    def __str__(self):
        return self.description()

    def description(
        self, domain_volume=True, surface_cavity_volume=True, center_cavity_volume=True
    ):
        s = "{}, frame {}, resolution {}".format(
            os.path.basename(self.filepath), self.frame + 1, self.resolution
        )
        if (
            surface_cavity_volume
            and self.surface_cavities is not None
            and self.atoms.volume is not None
        ):
            cavvolume = np.sum(self.surface_cavities.volumes)
            volpercent = 100 * cavvolume / self.atoms.volume.volume
            s += ", {:0.1f}% cavities (surface-based)".format(volpercent)
        if (
            center_cavity_volume
            and self.center_cavities is not None
            and self.atoms.volume is not None
        ):
            cavvolume = np.sum(self.center_cavities.volumes)
            volpercent = 100 * cavvolume / self.atoms.volume.volume
            s += ", {:0.1f}% cavities (center-based)".format(volpercent)
        if domain_volume and self.domains is not None and self.atoms.volume is not None:
            cavvolume = np.sum(self.domains.volumes)
            volpercent = 100 * cavvolume / self.atoms.volume.volume
            s += ", {:0.1f}% cavities (domains)".format(volpercent)
        return s

    # Properties to be compatible to the old CalculationResults
    @property
    def number_of_atoms(self):
        logger.warn("use of deprecated property")
        return self.atoms.number

    @property
    def atom_positions(self):
        logger.warn("use of deprecated property")
        return self.atoms.positions

    @property
    def atom_radii(self):
        logger.warn("use of deprecated property")
        return self.atoms.radii

    @property
    def number_of_domains(self):
        logger.warn("use of deprecated property")
        if self.domains is not None:
            return self.domains.number
        else:
            return None

    @property
    def domain_volumes(self):
        logger.warn("use of deprecated property")
        if self.domains is not None:
            return self.domains.volumes
        else:
            return None

    @property
    def domain_surface_areas(self):
        logger.warn("use of deprecated property")
        if self.domains is not None:
            return self.domains.surface_areas
        else:
            return None

    @property
    def domain_centers(self):
        logger.warn("use of deprecated property")
        if self.domains is not None:
            return self.domains.centers
        else:
            return None

    @property
    def domain_triangles(self):
        logger.warn("use of deprecated property")
        if self.domains is not None:
            return self.domains.triangles
        else:
            return None

    @property
    def number_of_multicavities(self):
        logger.warn("use of deprecated property")
        if self.surface_cavities is not None:
            return self.surface_cavities.number
        else:
            return None

    @property
    def multicavity_volumes(self):
        logger.warn("use of deprecated property")
        if self.surface_cavities is not None:
            return self.surface_cavities.volumes
        else:
            return None

    @property
    def multicavity_surface_areas(self):
        logger.warn("use of deprecated property")
        if self.surface_cavities is not None:
            return self.surface_cavities.surface_areas
        else:
            return None

    @property
    def multicavities(self):
        logger.warn("use of deprecated property")
        if self.surface_cavities is not None:
            return self.surface_cavities.multicavities
        else:
            return None

    @property
    def multicavity_triangles(self):
        logger.warn("use of deprecated property")
        if self.surface_cavities is not None:
            return self.surface_cavities.triangles
        else:
            return None

    @property
    def number_of_center_multicavities(self):
        logger.warn("use of deprecated property")
        if self.center_cavities is not None:
            return self.center_cavities.number
        else:
            return None

    @property
    def center_multicavity_volumes(self):
        logger.warn("use of deprecated property")
        if self.center_cavities is not None:
            return self.center_cavities.volumes
        else:
            return None

    @property
    def center_multicavity_surface_areas(self):
        logger.warn("use of deprecated property")
        if self.center_cavities is not None:
            return self.center_cavities.surface_areas
        else:
            return None

    @property
    def center_multicavities(self):
        logger.warn("use of deprecated property")
        if self.center_cavities is not None:
            return self.center_cavities.multicavities
        else:
            return None

    @property
    def center_multicavity_triangles(self):
        logger.warn("use of deprecated property")
        if self.center_cavities is not None:
            return self.center_cavities.triangles
        else:
            return None



import itertools as it
from .node_border_iterator import iterate_node_border


class MergeGroup(object):
    """
    Class that collects a group of homogenous nodes. Groups that are divided by the periodic border can be
    organized in subgroups.
    The object is intended to be used in two phases:

        1.  Collecting nodes: All nodes are collected in one subgroup. Multiple subgroups are not allowed in this phase.
        2.  Merging with other MergeGroup objects: Two objects can be merged resulting in one object with separate
            subgroups. When merging has been started, no additional nodes can be added.

    """

    class SharedAttributes(object):
        """
        Instances of this class can be used to store key value pairs (like in dicts). One instance can be shared between
        multiple MergeGroup instances so an attribute update will automtically be published to all merge groups that
        hold a reference to the same shared attributes object.
        """

        def __init__(self, **kwargs):
            for key, value in kwargs.items():
                setattr(self, key, value)

        def __contains__(self, item):
            return item in self.__dict__

    def __init__(self, initial_node=None):
        self._shared_attributes = MergeGroup.SharedAttributes(
            subgroups=[],
            translation_vectors=[],  # offset that can be applied to the i-th subgroup to merge it with the next
            # subgroup in the list
            all_merge_groups=[self],
            is_cyclic=False,
        )
        self._index_of_primary_subgroup = (
            None  # Saves the index of the first subgroup in the subgroups list that has
        )
        # been assigned to this MergeGroup object
        if initial_node is not None:
            self.add(set([initial_node]))

    def add(self, nodes):
        """
        Adds nodes from another MergeGroup or a single node.
        """
        if len(self._subgroups) > 1:
            raise AddingNodesNotAllowedError(
                "In the merge phase adding nodes is not longer possible."
            )

        if isinstance(nodes, type(self)):
            # only the nodes of the second merge group are collected. That is NOT a merging operation!
            self.add(list(nodes))
        else:
            if not isinstance(nodes, (set, list)):
                nodes = [nodes]
            if len(self._subgroups) == 0:
                self._subgroups.append(set())
                self._index_of_primary_subgroup = 0
            self._subgroups[0].update(nodes)

    def merge(self, merge_group, translation_vector):
        """
        Merges two MergeGroups by adding the _subgroups of the other merge_group respectively. Therefore, both
        MergeGroups are changed.
        It is supposed that the given translation_vector can be applied on the primary subgroup of self to merge it with
        the primary subgroup of merge_group.
        """

        def subtract_vector_list(translation_vector, *vector_lists):
            for vector in it.chain(*vector_lists):
                translation_vector = tuple(
                    tc - vc for tc, vc in zip(translation_vector, vector)
                )
            return translation_vector

        subgroup_count_before_merge = len(self._subgroups)

        self._subgroups.extend(merge_group._subgroups)
        subtraction_vectors = self._translation_vectors[
            self._index_of_primary_subgroup :
        ]
        subtraction_vectors.extend(
            merge_group._translation_vectors[: merge_group._index_of_primary_subgroup]
        )
        self._translation_vectors.append(
            subtract_vector_list(translation_vector, subtraction_vectors)
        )
        self._translation_vectors.extend(merge_group._translation_vectors)

        self._all_merge_groups.extend(merge_group._all_merge_groups)
        merge_group._update_all_merge_groups(
            self._shared_attributes, subgroup_count_before_merge
        )

    def set_cyclic(self):
        self._shared_attributes.is_cyclic = True

    def _update_all_merge_groups(self, shared_attributes, index_shift):
        for merge_group in self._all_merge_groups[:]:
            merge_group._shared_attributes = shared_attributes
            merge_group._index_of_primary_subgroup += index_shift

    def __getattr__(self, attr):
        if attr.startswith("_"):
            if (
                hasattr(self, "_shared_attributes")
                and attr[1:] in self._shared_attributes
            ):
                return getattr(self._shared_attributes, attr[1:])
        return super(MergeGroup, self).__getattribute__(attr)

    def __contains__(self, node):
        return any(node in subgroup for subgroup in self._subgroups)

    def __len__(self):
        return sum(len(subgroup) for subgroup in self._subgroups)

    def __delitem__(self, item):
        for subgroup in self._subgroups:
            if item in subgroup:
                del subgroup[item]

    def __iter__(self):
        """
        Iterator for the nodes of all subgroups 'as is' -> translation is not applied.
        """
        return it.chain(*(subgroup for subgroup in self._subgroups))

    def iter_with_applied_translation(
        self,
        iter_with_non_translated_nodes=False,
        keep_largest_volume_within_cell=False,
    ):
        """
        Iterator for the nodes of all subgroups with applied translation -> the result is a continuous volume
        """

        def create_node_generator(subgroup, translation_vector):
            # generator expression must be created in a nested function to avoid late variable binding problems with
            # "combined_translation_vector"; without a nested function all generator expressions would get the last
            # value that was assigned to "combined_translation_vector" during the loop iterations in the outer function
            return (
                (tuple(nc - tc for nc, tc in zip(pos, translation_vector)), dim)
                for pos, dim in subgroup
            )

        def get_generators_from_start_index(index, reverse=False):
            generators = []
            if not reverse:
                factor = 1
                subgroups = self._subgroups[index:]
                translation_vectors = self._translation_vectors[index:]
                combined_translation_vector = (0, 0, 0)
            else:
                factor = -1
                subgroups = self._subgroups[index - 1 :: -1] if index > 0 else []
                translation_vectors = (
                    self._translation_vectors[index - 2 :: -1] if index > 1 else []
                )
                combined_translation_vector = (
                    tuple(-c for c in self._translation_vectors[index - 1])
                    if index > 0
                    else (0, 0, 0)
                )
            iterator = it.zip_longest(subgroups, translation_vectors)
            for subgroup, translation_vector in iterator:
                generator = create_node_generator(subgroup, combined_translation_vector)
                if iter_with_non_translated_nodes:
                    generators.append(zip(subgroup, generator))
                else:
                    generators.append(generator)
                if translation_vector is not None:
                    combined_translation_vector = tuple(
                        cc + factor * tc
                        for cc, tc in zip(
                            combined_translation_vector, translation_vector
                        )
                    )
            return generators

        if keep_largest_volume_within_cell:
            index_of_non_translated_subgroup = max(
                range(len(self._subgroups)), key=lambda i: len(self._subgroups[i])
            )
        else:
            index_of_non_translated_subgroup = self._index_of_primary_subgroup
        generators = []
        backwards_generators = get_generators_from_start_index(
            index_of_non_translated_subgroup, reverse=True
        )
        generators.extend(list(reversed(backwards_generators)))
        forwards_generators = get_generators_from_start_index(
            index_of_non_translated_subgroup
        )
        generators.extend(forwards_generators)
        return it.chain(*generators)

    @property
    def is_cyclic(self):
        return self._is_cyclic


class Graph(object):
    def __init__(self):
        self.nodes = {}

    def add_node(self, node, neighbors=None):
        if neighbors is None:
            neighbors = set()
        else:
            neighbors = set(neighbors)

        if node in self.nodes:
            self.remove_node(node)

        self.nodes[node] = set()
        self.add_neighbors(node, neighbors)

    def add_neighbors(self, node, neighbors):
        for neighbor in neighbors:
            if neighbor in self.nodes:
                self.nodes[neighbor].add(node)
            else:
                self.add_node(neighbor, [node])
        self.nodes[node] |= set(neighbors)

    def remove_node(self, node):
        neighbors = self.nodes[node]
        self.remove_neighbors(node, self.nodes[node])
        del self.nodes[node]
        return neighbors

    def remove_neighbors(self, node, neighbors):
        for neighbor in neighbors:
            self.nodes[neighbor] -= set([node])
        self.nodes[node] -= set(neighbors)
        return self.nodes[node]

    def get_neighbors(self, node):
        return self.nodes[node]

    def __contains__(self, node):
        return node in self.nodes

    def __len__(self):
        return len(self.nodes)

    def __iter__(self):
        return self.nodes.__iter__()

    def iterkeys(self):
        return self.nodes.iterkeys()

    def items(self):
        return self.nodes.items()

    def itervalues(self):
        return self.nodes.itervalues()


class GraphForSplitAndMerge(Graph):
    """
    Nodes must be tuples of the format ((x, y, z), (width, height, depth)).
    Graph class which starts with just one initial node. That node can be split
    multiple times at a specified split point (resulting in max 8 sub nodes).
    Neighboring relationships of the sub nodes are determined automatically.
    After the split phase, boundary nodes can be detected. If periodic boundary
    conditions should be considered, the neighboring relationships can be updated
    manually with the found boundary nodes.
    Afterwards, single nodes can be merged together. Internally, they stay as
    single nodes and merges are logged with other data structures (sets).
    The split method can only be used until merge is called for the first time!
    """

    def __init__(
        self, data, mask, get_translation_vector, is_relevant_part, initial_node=None
    ):
        Graph.__init__(self)
        self.data = data
        self.mask = mask
        self.get_translation_vector = get_translation_vector
        self.is_relevant_part = is_relevant_part
        self.merged_nodes = {}
        self.border_nodes = None
        self.border_node_translation_vectors = None
        self.border_node_pair_translations = None
        self.split_allowed = True
        self.adding_non_periodic_neighbors_allowed = True
        self.merging_non_periodic_neighbors_allowed = True
        self.initial_node_set = False
        if initial_node is not None:
            self.set_initial_node(initial_node)

    def set_initial_node(self, initial_node):
        if not self.initial_node_set:
            self.add_node(initial_node, set())
            self.initial_node_set = True
        else:
            raise InitialNodeAlreadySetError

    def add_node(self, node, potential_neighbors=None):
        neighbors = self.__find_neighbors(node, potential_neighbors)
        Graph.add_node(self, node, neighbors)
        # The new node must be merged with itself because later other nodes could share the set of merged nodes:
        self.merged_nodes[node] = MergeGroup(node)

    def add_neighbors(self, node, neighbors, translation_vectors=None):
        periodic_border_relationship = translation_vectors is not None
        if not self.adding_non_periodic_neighbors_allowed:
            if node not in self.nodes:
                raise AddingNodesNotAllowedError
            elif not periodic_border_relationship:
                raise AddingNonPeriodicNeighborsNotAllowedError
        Graph.add_neighbors(self, node, neighbors)
        if periodic_border_relationship:
            self.adding_non_periodic_neighbors_allowed = False
            if self.border_nodes is None:
                self.__mark_border_nodes()
            self.border_nodes[node] |= set(neighbors)
            for neighbor, translation_vector in zip(neighbors, translation_vectors):
                self.border_node_pair_translations[(neighbor, node)] = (
                    translation_vector
                )
                self.border_node_pair_translations[(node, neighbor)] = tuple(
                    -c for c in translation_vector
                )

    def remove_node(self, node):
        Graph.remove_node(self, node)
        del self.merged_nodes[node]

    def split_node(self, node, split_point_rel):
        """
        split_point_rel is that point relative to the left top corner of the node that contains the first inhomogeneity.
        It is implicated that the node data is stored in C order. Therefore, it is possible to split the node into 4
        homogeneous and 4 potential inhomogeneous sub nodes which causes a great speedup of the whole
        algorithm.
        """
        if self.split_allowed:
            self.border_nodes = None
            self.border_node_translation_vectors = None
            self.border_node_pair_translations = None
            x, y, z = node[0]
            w, h, d = node[1]
            x_inh, y_inh, z_inh = split_point_rel

            potential_new_homogen_nodes = (
                ((x, y, z), (x_inh + 1, y_inh + 1, z_inh)),
                ((x, y, z + z_inh), (x_inh + 1, y_inh, d - z_inh)),
                ((x, y + y_inh + 1, z), (x_inh, h - y_inh - 1, z_inh)),
                ((x, y + y_inh, z + z_inh), (x_inh, h - y_inh, d - z_inh)),
            )

            potential_new_inhomogen_nodes = (
                ((x + x_inh + 1, y, z), (w - x_inh - 1, y_inh + 1, z_inh)),
                ((x + x_inh + 1, y, z + z_inh), (w - x_inh - 1, y_inh, d - z_inh)),
                ((x + x_inh, y + y_inh + 1, z), (w - x_inh, h - y_inh - 1, z_inh)),
                ((x + x_inh, y + y_inh, z + z_inh), (w - x_inh, h - y_inh, d - z_inh)),
            )

            def get_relevant_nodes(potential_nodes, hom_nodes=False):
                new_nodes = []
                for n in potential_nodes:
                    x, y, z = n[0]
                    w, h, d = n[1]
                    if w > 0 and h > 0 and d > 0:
                        # a => b
                        if not hom_nodes or (
                            self.is_relevant_part(
                                self.data[x : x + w, y : y + h, z : z + d]
                            )
                            and not bool(self.mask[x, y, z])
                        ):
                            new_nodes.append(n)
                return new_nodes

            new_homogen_nodes = get_relevant_nodes(
                potential_new_homogen_nodes, hom_nodes=True
            )
            new_inhomogen_nodes = get_relevant_nodes(potential_new_inhomogen_nodes)
            all_new_nodes = set(new_homogen_nodes) | set(new_inhomogen_nodes)

            potential_neighbors = self.get_neighbors(node) | all_new_nodes
            self.remove_node(node)
            for n in all_new_nodes:
                self.add_node(n, potential_neighbors - set([n]))

            return new_inhomogen_nodes
        else:
            raise SplitNotAllowedError

    def get_border_nodes(self):
        if self.border_nodes is None:
            self.__mark_border_nodes()
        return self.border_nodes.keys()

    def get_border_node_translation_vectors(self):
        if self.border_node_translation_vectors is None:
            self.__mark_border_nodes()
        return self.border_node_translation_vectors

    def __mark_border_nodes(self):
        translation_vectors = None

        def func(border_x, border_y, border_z):
            try:
                if bool(self.mask[border_x, border_y, border_z]):
                    translation_vectors.add(
                        tuple(
                            self.get_translation_vector((border_x, border_y, border_z))
                        )
                    )
            except:
                pass  # TODO dont fix like this

        self.border_nodes = {}
        self.border_node_translation_vectors = {}
        self.border_node_pair_translations = {}
        for node in self:
            node_x, node_y, node_z = node[0]
            node_w, node_h, node_d = node[1]
            for x, y, z in it.product(
                (node_x - 1, node_x + node_w),
                (node_y - 1, node_y + node_h),
                (node_z - 1, node_z + node_d),
            ):
                if bool(self.mask[x, y, z]):
                    translation_vectors = set()
                    iterate_node_border(node, func)
                    self.border_nodes[node] = set()
                    self.border_node_translation_vectors[node] = translation_vectors
                    break

    def forbid_splitting(self):
        self.split_allowed = False

    def merge_nodes(self, node1, node2):
        """
        Only nodes can be merged that are neighboring.
        """

        # Neighbors?
        if node2 not in self.nodes[node1]:
            raise NotNeighboringError
        self.forbid_splitting()
        if not self.is_merged(node1, node2, detect_cyclic_merge=True):
            separated_by_periodic_boundary_condition = (
                self.border_nodes is not None
                and node1 in self.border_nodes
                and node2 in self.border_nodes[node1]
            )
            if not separated_by_periodic_boundary_condition:
                if not self.merging_non_periodic_neighbors_allowed:
                    raise MergingNonBorderNodesNotAllowedError
                # Record all merged nodes of node2 as merged nodes of node1 (node2 is already included)
                self.merged_nodes[node1].add(self.merged_nodes[node2])
                # Since the merged nodes of node1 have a reference to self.merged_nodes[node1], they have been already
                # updated. So, it is only necessary to update the merged_nodes of node2 (which contain node2 already).
                for merged_node in self.merged_nodes[node2]:
                    self.merged_nodes[merged_node] = self.merged_nodes[node1]
            else:
                self.merging_non_periodic_neighbors_allowed = False
                self.merged_nodes[node1].merge(
                    self.merged_nodes[node2],
                    self.border_node_pair_translations[(node1, node2)],
                )
        return node2 in self.merged_nodes[node1]

    def is_merged(self, node1, node2, detect_cyclic_merge=False):
        # If node2 is not merged with node1 than node1 is not merged with node2, neither.
        is_merged = node2 in self.merged_nodes[node1]
        if (
            is_merged
            and detect_cyclic_merge
            and self.merged_nodes[node1] is self.merged_nodes[node2]
        ):
            self.merged_nodes[node1].set_cyclic()
        return is_merged

    def get_all_areas(
        self,
        apply_translation=False,
        with_non_translated_nodes=False,
        mark_cyclic_areas=False,
    ):
        """
        Returns a list of sets, each describing a merged area of nodes. If "apply_translation" is set to True,
        all subgroups of an area are translated with respect to the periodic border condition. As a result, all areas
        are contiguous in space. If "with_non_translated_nodes" is set additionally, a second list is returned that
        contains the same areas without applied translations. If "mark_cyclic_areas" is set to True, the function has a
        further list as second/third return value indicating which areas (given by index) are cyclic (have infinite
        extent).
        """
        if self.split_allowed:
            return []
        areas = []
        # Saves non translated areas if "apply_translation" and "with_non_translated_nodes" are set to true
        alt_areas = [] if with_non_translated_nodes else None
        cyclic_areas = [] if mark_cyclic_areas else None
        visited_nodes = set()
        for node in self:
            if node not in visited_nodes:
                area = set()
                if apply_translation:
                    node_iterator = self.merged_nodes[
                        node
                    ].iter_with_applied_translation(
                        iter_with_non_translated_nodes=True,
                        keep_largest_volume_within_cell=True,
                    )
                    if with_non_translated_nodes:
                        alt_area = set()
                        for merged_node, translated_node in node_iterator:
                            area.add(translated_node)
                            alt_area.add(merged_node)
                            visited_nodes.add(merged_node)
                        alt_areas.append(alt_area)
                    else:
                        for merged_node, translated_node in node_iterator:
                            area.add(translated_node)
                            visited_nodes.add(merged_node)
                else:
                    node_iterator = self.merged_nodes[node]
                    for merged_node in node_iterator:
                        area.add(merged_node)
                        visited_nodes.add(merged_node)
                areas.append(area)
                if mark_cyclic_areas and self.merged_nodes[node].is_cyclic:
                    cyclic_areas.append(len(areas) - 1)
        if apply_translation and with_non_translated_nodes and mark_cyclic_areas:
            return areas, alt_areas, cyclic_areas
        elif mark_cyclic_areas:
            return areas, cyclic_areas
        else:
            return areas

    def __find_neighbors(self, node, potential_neighbors):
        x_, y_, z_ = node[0]
        w_, h_, d_ = node[1]
        neighbors = set()
        if potential_neighbors is None:
            potential_neighbors = self.nodes.keys()
        for n in potential_neighbors:
            x, y, z = n[0]
            w, h, d = n[1]
            if (
                x - w_ <= x_ <= x + w
                and y - h_ <= y_ <= y + h
                and z - d_ <= z_ <= z + d
            ):
                neighbors.add(n)
        return neighbors

    def iter_border_items(self):
        self.get_border_nodes()  # ensure that self.border_nodes is set correctly
        return self.border_nodes.items()


class NotNeighboringError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class InitialNodeAlreadySetError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class SplitNotAllowedError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class AddingNodesNotAllowedError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class AddingNonPeriodicNeighborsNotAllowedError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class MergingNonBorderNodesNotAllowedError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)



# -*- coding: utf-8 -*-

from PySide6 import QtCore, QtWidgets


class TableFit:
    DEFAULT_SCROLLBAR_EXTRA_SPACE = (16, 16)

    def __init__(self, table_widget, size_hint_only=False, scrollbar_extra_space=None):
        self._table_widget = table_widget
        self._table_widget.sizeHint = self.sizeHint
        if scrollbar_extra_space is None:
            self._scrollbar_extra_space = TableFit.DEFAULT_SCROLLBAR_EXTRA_SPACE
        else:
            self._scrollbar_extra_space = []
            for given_value, default_value in zip(
                scrollbar_extra_space, TableFit.DEFAULT_SCROLLBAR_EXTRA_SPACE
            ):
                value = given_value if given_value >= 0 else default_value
                self._scrollbar_extra_space.append(value)
            self._scrollbar_extra_space = tuple(self._scrollbar_extra_space)

        if not size_hint_only:
            self._table_widget.setSizePolicy(
                QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum
            )
            self._table_widget.setHorizontalScrollBarPolicy(
                QtCore.Qt.ScrollBarAlwaysOff
            )
            self._table_widget.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)

    def sizeHint(self):
        horizontal_header = self._table_widget.horizontalHeader()
        vertical_header = self._table_widget.verticalHeader()
        width = (
            horizontal_header.length()
            + vertical_header.width()
            + self._scrollbar_extra_space[0]
        )
        height = (
            vertical_header.length()
            + horizontal_header.height()
            + self._scrollbar_extra_space[1]
        )
        return QtCore.QSize(width, height)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from PySide6 import QtCore, QtWidgets


class CutoffTableWidget(QtWidgets.QTableWidget):
    class CutoffLineEdit(QtWidgets.QLineEdit):
        focus_in = QtCore.Signal()

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def focusInEvent(self, event):
            super(CutoffTableWidget.CutoffLineEdit, self).focusInEvent(event)
            self.focus_in.emit()

    text_edited = QtCore.Signal()

    def __init__(self, radii, parent=None):
        super().__init__(len(radii), 2, parent)
        self._radii = radii
        self._init_ui()

    def _init_ui(self):
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        self.setHorizontalHeaderLabels(("Covalent Radius", "Cutoff Radius"))
        self.setVerticalHeaderLabels(self._radii.keys())
        for i in range(len(self._radii)):
            self.setItem(i, 0, QtWidgets.QTableWidgetItem())
            self.item(i, 0).setText(str(list(self._radii.values())[i]))
            current_line_edit = CutoffTableWidget.CutoffLineEdit()
            current_line_edit.textEdited.connect(self.text_edited)
            current_line_edit.focus_in.connect(
                lambda row_index=i: self._line_edit_focus_in(row_index)
            )
            self.setCellWidget(i, 1, current_line_edit)
        self.setShowGrid(True)

    def _line_edit_focus_in(self, row_index):
        self.setCurrentIndex(self.model().index(row_index, 1))

    def keyPressEvent(self, event):
        key_code = event.key()
        if key_code == QtCore.Qt.Key_Tab or key_code == QtCore.Qt.Key_Backtab:
            index_increment = 1 if key_code == QtCore.Qt.Key_Tab else -1
            row_index = (
                self.currentIndex().row() + index_increment + len(self._radii)
            ) % len(self._radii)
            self.setCurrentIndex(self.model().index(row_index, 1))
        else:
            super(CutoffTableWidget, self).keyPressEvent(event)



from PySide6 import QtCore, QtWidgets
from collections import Counter, OrderedDict


class TreeList(QtWidgets.QTreeWidget):
    def __init__(self, html_view):
        QtWidgets.QTreeWidget.__init__(self)
        self.setColumnCount(1)
        # todo dynamic list generation from results

        # attributes get intialized by update_control method after calculation
        self.control = None
        self.atoms = None
        self.atom_number = None
        self.atom_elements = {}
        self.cavities_center = None
        self.cavities_surface = None
        self.domains = None

        self.atom_list = []
        self.cavities_center_list = []
        self.cavities_surface_list = []
        self.domains_list = []

        self.html_view = html_view

        self.items = []
        self.update_tree_view()

        # expand all items per default
        for item in self.items:
            item.setExpanded(True)

        self.adjustSize()
        self.updateGeometry()
        self.resizeColumnToContents(0)
        self.addTopLevelItems(self.items)
        self.setHeaderHidden(True)
        self.itemSelectionChanged.connect(self.item_selection_changed)

    def item_selection_changed(self):
        if self.selectedItems():
            selected_item = self.selectedItems()[0]
            self.item_selected(selected_item, 0)

    def select_atom(self, index):
        self.setCurrentItem(self.items[0].child(index))

    def select_center_cavity(self, index):
        self.setCurrentItem(self.items[1].child(index))

    def select_surface_cavity(self, index):
        self.setCurrentItem(self.items[2].child(index))

    def select_domain(self, index):
        self.setCurrentItem(self.items[3].child(index))

    def item_selected(self, item, column):
        """
        This method decides which element in the tree_list was clicked and calls the specific show method.
        """
        data = item.data(0, column)
        if data == "Atoms":
            self.html_view.show_atom_group()
        elif data.startswith("Atom "):
            index = int(data.split()[-1]) - 1
            self.html_view.show_atom(index)
        elif data == "Cavities (center)":
            self.html_view.show_center_cavity_group()
        elif data == "Cavities (surface)":
            self.html_view.show_surface_cavity_group()
        elif data == "Cavities (domains)":
            self.html_view.show_domain_group()
        elif item.parent().data(0, column) is not None:
            parent = item.parent().data(0, column)
            if data.startswith("Cavity ") and parent == "Cavities (center)":
                index = int(data.split()[-1]) - 1
                self.html_view.show_center_cavity(index)
            elif data.startswith("Cavity ") and parent == "Cavities (surface)":
                index = int(data.split()[-1]) - 1
                self.html_view.show_surface_cavity(index)
            elif data.startswith("Cavity ") and parent == "Cavities (domains)":
                index = int(data.split()[-1]) - 1
                self.html_view.show_domain(index)

    def update_results(self, results):
        self.atoms = results.atoms
        self.atom_number = self.atoms.number
        self.atom_elements = Counter(self.atoms.elements)
        self.cavities_center = results.center_cavities
        self.cavities_surface = results.surface_cavities
        self.domains = results.domains

        self.atom_list = ["Atom %d" % (i + 1) for i in range(self.atoms.number)]
        if self.cavities_center is not None:
            self.cavities_center_list = [
                "Cavity %d" % (i + 1)
                for i in range(len(self.cavities_center.multicavities))
            ]
        else:
            self.cavities_center_list = []
        if self.cavities_surface is not None:
            self.cavities_surface_list = [
                "Cavity %d" % (i + 1)
                for i in range(len(self.cavities_surface.multicavities))
            ]
        else:
            self.cavities_surface_list = []
        if self.domains is not None:
            self.domains_list = [
                "Cavity %d" % (i + 1) for i in range(self.domains.number)
            ]
        else:
            self.domains_list = []
        # print dir(atoms)
        # print atoms.positions
        # print atoms.elements
        # print atoms.sorted_positions
        # print atoms.radii_as_indices
        # print atoms.sorted_radii
        # print atoms.volume

        self.update_tree_view()

    def update_tree_view(self):
        for index in reversed(range(self.topLevelItemCount())):
            self.takeTopLevelItem(index)
        src = OrderedDict(
            [
                ("Atoms", self.atom_list),
                ("Cavities (center)", self.cavities_center_list),
                ("Cavities (surface)", self.cavities_surface_list),
                ("Cavities (domains)", self.domains_list),
            ]
        )

        self.items = []
        for root, sib in src.items():
            self.items.append(QtWidgets.QTreeWidgetItem(self, [root]))
            if sib:
                self.items[-1].addChildren(
                    [QtWidgets.QTreeWidgetItem(self.items[-1], [s]) for s in sib]
                )



# -*- coding: utf-8 -*-


from __future__ import absolute_import


from PySide6 import QtCore, QtWidgets
from .util.table_with_removeable_items import TableWithRemoveableEntries
from .util.cutoff_preview_table import CutoffPreviewTable


class CutoffHistoryTable(TableWithRemoveableEntries):
    HEADER_LABELS = ("Filename", "Frame", "Time", "Cutoff")

    def __init__(self, history):
        self._history = list(history)
        super().__init__(self.HEADER_LABELS, len(history))
        self._orig_index_to_row_index = range(len(self._history))
        self._indices_to_remove = set()
        self._init_ui()

    def _init_ui(self):
        pass

    @property
    def history(self):
        return [
            entry
            for i, entry in enumerate(self._history)
            if i not in self._indices_to_remove
        ]

    @property
    def history_entries_for_deletion(self):
        return [self._history[i] for i in sorted(self._indices_to_remove)]

    def entry_count(self):
        return len(self._history)

    def entries(self):
        entries = []
        for i, history_entry in enumerate(self._history):
            current_row = []
            for j, value in enumerate(
                (history_entry.filename, history_entry.frame + 1, history_entry.time)
            ):
                item = QtWidgets.QTableWidgetItem(str(value))
                item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
                current_row.append(item)
            cutoff_table = CutoffPreviewTable(history_entry.radii)
            current_row.append(cutoff_table)
            entries.append(tuple(current_row))
        return entries



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import os
import sys

from ..core import bonds
from ..core import control

from .tabs.file_tab import FileTabDock
from .tabs.view_tab import ViewTabDock
from .tabs.image_video_tab import ImageVideoTabDock
from .tabs.statistics_tab import StatisticsTabDock
from .tabs.log_tab import LogTabDock
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtQuick import QQuickWindow, QSGRendererInterface
from .dialogs.settings_dialog import SettingsDialog
from .dialogs.about_dialog import AboutDialog
from .gl_widget import UpdateGLEvent
from ..util import message
from .gl_stack import GLStack
from .._version import __version__
import functools
import os.path
from ..config.configuration import config


WEBSITE_URL = "https://pgi-jcns.fz-juelich.de/portal/pages/pymoldyn-doc.html"


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, control):
        QtWidgets.QMainWindow.__init__(self, None)
        QQuickWindow.setGraphicsApi(QSGRendererInterface.GraphicsApi.OpenGL)

        self.control = control
        self.center = CentralWidget(self)
        self.file_dock = FileTabDock(self)
        self.view_dock = ViewTabDock(self)
        self.view_dock.setVisible(False)
        self.image_video_dock = ImageVideoTabDock(self)
        self.statistics_dock = StatisticsTabDock(self)
        self.log_dock = LogTabDock(self)
        self._error_wrapper = None

        p = self.file_dock.file_tab.progress_dialog
        self.set_output_callbacks(
            p.progress,
            p.print_step,
            p.calculation_finished,
            self.show_error,
            self.log_dock.append_log,
        )

        self.docks = []

        self.shown_dataset = None

        self.setTabPosition(QtCore.Qt.RightDockWidgetArea, QtWidgets.QTabWidget.North)

        self.setCentralWidget(self.center)

        for dock in (
            self.file_dock,
            self.view_dock,
            self.image_video_dock,
            self.statistics_dock,
            self.log_dock,
        ):
            self.addDockWidget(QtCore.Qt.RightDockWidgetArea, dock, QtCore.Qt.Vertical)
            self.docks.append(dock)

        for dock in (self.view_dock, self.statistics_dock):
            self.tabifyDockWidget(self.file_dock, dock)

        # this variable is used to open the FileDialog in the propper path
        self.file_dock.file_tab.most_recent_path = "~"

        self.menubar = None
        self.file_menu = None
        self.recent_files_submenu = None
        self.init_menu()

        self.setWindowTitle("pyMolDyn v%s" % __version__)
        self.setWindowIcon(QtGui.QIcon("icon.png"))

        self.show()
        # get Dock Widgets TabBar and set the first one to current
        self.file_dock.show()
        self.file_dock.raise_()

        # another workaround to do the same
        # tabbars = self.findChildren(QtWidgets.QTabBar)
        # tabbars[0].setCurrentIndex(0)

    def init_menu(self):
        open_action = QtGui.QAction("&Open dataset", self)
        open_action.setShortcut("Ctrl+O")
        open_action.triggered.connect(self.file_dock.file_tab.open_file_dialog)

        settings_action = QtGui.QAction("&Settings", self)
        settings_action.setShortcut("Ctrl+I")
        settings_action.triggered.connect(self.show_settings)

        export_submenu = QtWidgets.QMenu("&Export", self)

        export_bonds_action = QtGui.QAction("Export &Bonds", self)
        export_bonds_action.setShortcut("Ctrl+1")
        export_bonds_action.triggered.connect(self.wrapper_export_bonds)

        export_bond_angles_action = QtGui.QAction("Export Bond &Angles", self)
        export_bond_angles_action.setShortcut("Ctrl+2")
        export_bond_angles_action.triggered.connect(self.wrapper_export_bond_angles)

        export_bond_dihedral_angles_action = QtGui.QAction(
            "Export Bond &Dihedral Angles", self
        )
        export_bond_dihedral_angles_action.setShortcut("Ctrl+3")
        export_bond_dihedral_angles_action.triggered.connect(
            self.wrapper_export_bond_dihedral_angles
        )

        export_domains_action = QtGui.QAction(
            "Export Cavity Information (domains)", self
        )
        export_domains_action.setShortcut("Ctrl+4")
        export_domains_action.triggered.connect(self.wrapper_export_domains)

        export_surface_cavities_action = QtGui.QAction(
            "Export Cavity Information (surface method)", self
        )
        export_surface_cavities_action.setShortcut("Ctrl+5")
        export_surface_cavities_action.triggered.connect(
            self.wrapper_export_surface_cavities
        )

        export_center_cavities_action = QtGui.QAction(
            "Export Cavity Information (center method)", self
        )
        export_center_cavities_action.setShortcut("Ctrl+6")
        export_center_cavities_action.triggered.connect(
            self.wrapper_export_center_cavities
        )

        website_action = QtGui.QAction("&pyMolDyn website", self)
        website_action.setShortcut("F1")
        website_action.triggered.connect(self.show_website)

        about_action = QtGui.QAction("&About", self)
        about_action.triggered.connect(self.show_about_box)

        self.menubar = self.menuBar()
        self.file_menu = self.menubar.addMenu("&File")
        self.file_menu.addAction(open_action)
        self.file_menu.addAction(settings_action)
        self.file_menu.addMenu(export_submenu)

        self.init_submenu_recent_files()
        self.file_menu.addMenu(self.recent_files_submenu)

        export_submenu.addAction(export_bonds_action)
        export_submenu.addAction(export_bond_angles_action)
        export_submenu.addAction(export_bond_dihedral_angles_action)
        export_submenu.addAction(export_domains_action)
        export_submenu.addAction(export_surface_cavities_action)
        export_submenu.addAction(export_center_cavities_action)

        help_menu = self.menubar.addMenu("&Help")
        help_menu.addAction(website_action)
        help_menu.addSeparator()
        help_menu.addAction(about_action)

    def show_error(self, error_message):
        QtCore.QMetaObject.invokeMethod(
            self,
            "_show_error",
            QtCore.Qt.QueuedConnection,
            QtCore.Q_ARG(str, error_message),
        )

    @QtCore.Slot(str)
    def _show_error(self, error_message):
        QtWidgets.QMessageBox.information(self, "Information", error_message)

    def show_settings(self):
        SettingsDialog()
        self.control.update()
        self.statistics_dock.update_results(self.control.visualization.results)

    def show_website(self):
        url = QtCore.QUrl(WEBSITE_URL)
        QtGui.QDesktopServices.openUrl(url)

    def show_about_box(self):
        AboutDialog(
            self,
            "pyMolDyn is a molecule viewer which is capable of computing molecular cavities.",
            (
                ("Florian Rhiem", "f.rhiem@fz-juelich.de"),
                ("Fabian Beule", "f.beule@fz-juelich.de"),
                ("David Knodt", "d.knodt@fz-juelich.de"),
                ("Ingo Meyer", "i.meyer@fz-juelich.de"),
                ("Torben Moll", "t.moll@fz-juelich.de"),
                ("Florian Macherey", "f.macherey@fz-juelich.de"),
            ),
        ).show()

    def init_submenu_recent_files(self):
        self.recent_files_submenu = QtWidgets.QMenu("&Recent files", self)
        if (not config.recent_files) or (config.recent_files == [""]):
            self.recent_files_submenu.setDisabled(True)
        else:
            self.recent_files_submenu.setEnabled(True)
            for f in config.recent_files:
                f_action = QtGui.QAction(f, self)
                f_action.triggered.connect(
                    functools.partial(self.wrapper_recent_files, f)
                )
                self.recent_files_submenu.addAction(f_action)

            self.file_dock.file_tab.most_recent_path = os.path.dirname(
                config.recent_files[0]
            )
            self._submenu_add_shortcut_for_first_item()

    def update_submenu_recent_files(self):
        if not config.recent_files:
            return

        most_recent_file = config.recent_files[0]
        if not most_recent_file:
            return

        actions_in_menu = self.recent_files_submenu.actions()
        actions_in_menu_str = [s.text() for s in actions_in_menu]

        if most_recent_file in actions_in_menu_str:
            index = actions_in_menu_str.index(most_recent_file)
            if index == 0:
                return

            self.recent_files_submenu.removeAction(actions_in_menu[index])
            self.recent_files_submenu.insertAction(
                actions_in_menu[0], actions_in_menu[index]
            )
        else:
            new_action = QtGui.QAction(most_recent_file, self)
            new_action.triggered.connect(
                functools.partial(self.wrapper_recent_files, most_recent_file)
            )

            if not actions_in_menu:
                self.recent_files_submenu.setEnabled(True)
                self.recent_files_submenu.addAction(new_action)
            else:
                self.recent_files_submenu.insertAction(actions_in_menu[0], new_action)
                self.recent_files_submenu.actions()[0].setDisabled(True)

                if len(actions_in_menu) == 5:
                    self.recent_files_submenu.removeAction(actions_in_menu[4])

        self.file_dock.file_tab.most_recent_path = os.path.dirname(most_recent_file)
        self._submenu_add_shortcut_for_first_item()
        self.recent_files_submenu.update()

    def _submenu_add_shortcut_for_first_item(self):
        actions_in_menu = self.recent_files_submenu.actions()
        actions_in_menu[0].setShortcut("Alt+1")
        for action in actions_in_menu[1:]:
            action.setShortcut("")
        self.recent_files_submenu.update()

    def wrapper_recent_files(self, f):
        if f:
            self.file_dock.file_tab.disable_files_in_menu_and_open(f)
            self.update_submenu_recent_files()

    def wrapper_export_bonds(self):
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export Bonds", "bonds.txt"
        )[0]
        if filename:
            core.bonds.export_bonds(filename, self.control.visualization.results.atoms)
            QtWidgets.QMessageBox.information(
                self, "Export Bonds", "Saved to filename: %s" % (filename)
            )

    def wrapper_export_bond_angles(self):
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export Bond Angles", "bond_angles.txt"
        )[0]
        if filename:
            core.bonds.export_bond_angles(
                filename, self.control.visualization.results.atoms
            )
            QtWidgets.QMessageBox.information(
                self, "Export Bond Angles", "Saved to filename: %s" % (filename)
            )

    def wrapper_export_bond_dihedral_angles(self):
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export Bond Dihedral Angles", "bond_dihedral_angles.txt"
        )[0]
        if filename:
            core.bonds.export_bond_dihedral_angles(
                filename, self.control.visualization.results.atoms
            )
            QtWidgets.QMessageBox.information(
                self,
                "Export Bond Dihedral Angles",
                "Saved to filename: %s" % (filename),
            )

    def wrapper_export_domains(self):
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export Cavity Information (domains)", "domains"
        )[0]
        if filename:
            filenames = self.control.visualization.results.domains.totxt(
                filename + "_{property}.txt"
            )
            QtWidgets.QMessageBox.information(
                self,
                "Export Cavity Information (domains)",
                "Saved to filenames: %s" % (", ".join(filenames)),
            )

    def wrapper_export_surface_cavities(self):
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export Cavity Information (surface method)", "surface_cavities"
        )[0]
        if filename:
            filenames = self.control.visualization.results.surface_cavities.totxt(
                filename + "_{property}.txt"
            )
            QtWidgets.QMessageBox.information(
                self,
                "Export Cavity Information (surface method)",
                "Saved to filenames: %s" % (", ".join(filenames)),
            )

    def wrapper_export_center_cavities(self):
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export Cavity Information (center method)", "center_cavities"
        )[0]
        if filename:
            filenames = self.control.visualization.results.center_cavities.totxt(
                filename + "_{property}.txt"
            )
            QtWidgets.QMessageBox.information(
                self,
                "Export Cavity Information (center method)",
                "Saved to filenames: %s" % (", ".join(filenames)),
            )

    def keyPressEvent(self, e):
        if e.key() == QtCore.Qt.Key_M:
            if not self.isFullScreen():
                for dock in self.docks:
                    dock.hide()
                self.showFullScreen()
            else:
                for dock in self.docks:
                    dock.show()
                self.showNormal()

    def set_output_callbacks(
        self, progress_func, print_func, finish_func, error_func, log_func
    ):
        message.set_output_callbacks(
            progress_func, print_func, finish_func, error_func, log_func
        )

    def updatestatus(self, was_successful=lambda: True):
        if was_successful and self.control.results is not None:
            results = self.control.results[-1][-1]
            self.shown_dataset = results
            visualization_settings = self.control.visualization.settings
            status = results.description(
                domain_volume=visualization_settings.show_domains,
                surface_cavity_volume=visualization_settings.show_surface_cavities,
                center_cavity_volume=visualization_settings.show_center_cavities,
            )
            self.statusBar().showMessage(status)
            self.statistics_dock.update_results(self.control.visualization.results)
            self.view_dock.setVisible(True)
            self.view_dock.view_tab.update_cavity_buttons(
                self.control.visualization.results, None
            )
            self.center.gl_stack.updatestatus()
            QtWidgets.QApplication.postEvent(
                self.center.gl_stack.gl_widget, UpdateGLEvent()
            )


#    def closeEvent(self, event):
#        reply = QtWidgets.QMessageBox.question(self, 'Message',
#            "Are you sure to quit?", QtWidgets.QMessageBox.Yes |
#            QtWidgets.QMessageBox.No, QtWidgets.QMessageBox.No)
#        if reply == QtWidgets.QMessageBox.Yes:
#            event.accept()
#        else:
#            event.ignore() TODO: why is this commented out


class CentralWidget(QtWidgets.QWidget):

    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self, parent)
        self.control = parent.control
        self.setWindowTitle("pyMolDyn 2")
        self.widget_titles = (
            "3D View",
            "Pair Distribution Functions",
            "Cavity Histograms",
        )
        self.init_gui()

    def init_gui(self):
        self.gl_stack = GLStack(self, self.parent())
        self.gl_widget = self.gl_stack.gl_widget
        self.combo = QtWidgets.QComboBox()
        for title in self.widget_titles:
            self.combo.addItem(title)
        self.combo.activated.connect(self.on_combo)  # removed [str] from activated

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.gl_stack)
        layout.addWidget(self.combo)

        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setLayout(layout)

    def on_combo(self, index):
        self.gl_stack.activate(index)



from PySide6 import QtCore, QtWidgets


class ProgressDialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.finished = False

        self.initUI()

    def initUI(self):
        hbox = QtWidgets.QVBoxLayout()
        self.label = QtWidgets.QLabel("starting calculation", self)
        self.progressbar = QtWidgets.QProgressBar(self)
        # self.cancel_btn = QtWidgets.QPushButton('Cancel', self)
        # self.cancel_btn.clicked.connect(self.cancel)
        self.setMinimumSize(300, 100)

        vbox = QtWidgets.QHBoxLayout()

        vbox.addStretch()
        # vbox.addWidget(self.cancel_btn)
        vbox.addStretch()

        hbox.addWidget(self.label)
        hbox.addWidget(self.progressbar)
        hbox.addLayout(vbox)

        self.setLayout(hbox)

    def calculation_finished(self):
        self.finished = True
        QtCore.QMetaObject.invokeMethod(
            self, "close_dialog", QtCore.Qt.QueuedConnection
        )

    def progress(self, value):
        QtCore.QMetaObject.invokeMethod(
            self.progressbar,
            "setValue",
            QtCore.Qt.QueuedConnection,
            QtCore.Q_ARG(int, value),
        )

    @QtCore.Slot()
    def close_dialog(self):
        if self.finished:
            self.done(QtWidgets.QDialog.Accepted)
        else:
            self.done(QtWidgets.QDialog.Rejected)
        pass

    def print_step(self, *text):
        tmp = str(text[0])
        for t in text[1:]:
            tmp += " {}".format(t)
        self.label.setText(tmp)

    def cancel(self):
        print("canceled calculation")
        self.close_dialog()



# -*- coding: utf-8 -*-

import os.path
from PySide6 import QtCore
from PySide6.QtWebEngineWidgets import QWebEngineView
from PySide6.QtWebEngineCore import (
    QWebEnginePage,
)  # QtWebKit is deprecated since Pyside2

has_qtwebengine = True


_js_inject_css_template = """\
window.onload = function() {{
    (function(css) {{
        var node = document.createElement('style');
        node.innerHTML = css;
        document.body.appendChild(node);
    }})('{css}');
}};\
"""

if has_qtwebengine:

    class WebWidget(QWebEngineView):
        gui_link_clicked = QtCore.Signal(str)

        class WebPage(QWebEnginePage):
            gui_link_clicked = QtCore.Signal(str)

            def __init__(self, html, css=None, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self._css = css
                self.setHtml(html)

            def setHtml(self, html):
                if html:
                    css = " ".join(self._css.split("\n"))
                    js = "<script>{code}</script>".format(
                        code=_js_inject_css_template.format(css=css)
                    )
                    html += js
                super(WebWidget.WebPage, self).setHtml(html)

            def acceptNavigationRequest(self, url, t, is_main_frame):
                if (
                    t == QWebEnginePage.NavigationTypeLinkClicked
                    and url.scheme() == "gui"
                ):
                    self.gui_link_clicked.emit(url.path())
                    return False
                else:
                    return True

        def __init__(self, css_filepath=None, *args, **kwargs):
            super().__init__(*args, **kwargs)
            if css_filepath is None:
                self._css = None
            else:
                css_filepath = os.path.join(os.path.dirname(__file__), css_filepath)
                with open(css_filepath) as f:
                    self._css = f.read()
            self._webpage = None

        def set_gui_html(self, html_string):
            if html_string is None:
                html_string = ""
            if self._webpage is None:
                self._webpage = WebWidget.WebPage(html_string, self._css, parent=self)
                self._webpage.gui_link_clicked.connect(
                    self.gui_link_clicked.emit
                )  # Forward signal
            else:
                self._webpage.setHtml(html_string)
            self.setPage(self._webpage)

else:

    class WebWidget(QWebView):
        gui_link_clicked = QtCore.Signal(str)

        class WebPage(QWebPage):
            gui_link_clicked = QtCore.Signal(str)

            def __init__(self, html, css, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self._css = css
                self.setLinkDelegationPolicy(QWebPage.DelegateAllLinks)
                self.linkClicked.connect(self._link_clicked)
                self.setHtml(html)

            def setHtml(self, html):
                if html:
                    css = " ".join(self._css.split("\n"))
                    js = "<script>{code}</script>".format(
                        code=_js_inject_css_template.format(css=css)
                    )
                    html += js
                self.mainFrame().setHtml(html)

            def _link_clicked(self, data):
                value = data.toString()
                if value.startswith("gui:"):
                    self.gui_link_clicked.emit(value[4:])

        def __init__(self, css_filepath=None, *args, **kwargs):
            super().__init__(*args, **kwargs)
            if css_filepath is None:
                self._css = None
            else:
                with open(css_filepath) as f:
                    self._css = f.read()
            self._webpage = None

        def set_gui_html(self, html_string):
            if html_string is None:
                html_string = ""
            if self._webpage is None:
                self._webpage = WebWidget.WebPage(html_string, self._css, parent=self)
                self._webpage.gui_link_clicked.connect(
                    self.gui_link_clicked.emit
                )  # Forward signal
            else:
                self._webpage.setHtml(html_string)
            self.setPage(self._webpage)



# -*- coding: utf-8 -*-

from PySide6 import QtCore, QtGui, QtWidgets
import datetime
import logging


STYLE_SHEET = """
.default {
    color: #909090;
    font-weight: bold;
}
.info {
    color: blue;
    font-weight: bold;
}
.warning {
    color: #DAA520;
    font-weight: bold;
}
.error {
    color: red;
    font-weight: bold;
}
"""


class LogTabDock(QtWidgets.QDockWidget):
    """
    DockWidget for the 'log'-tab
    """

    def __init__(self, parent):
        QtWidgets.QDockWidget.__init__(self, "logging", parent)
        self.setWidget(QtWidgets.QWidget())

        self.layout = QtWidgets.QHBoxLayout()
        self.log_tab = LogTab(self.widget(), parent)

        self.layout.addWidget(self.log_tab)
        self.widget().setLayout(self.layout)

        self.setFeatures(
            QtWidgets.QDockWidget.DockWidgetMovable
            | QtWidgets.QDockWidget.DockWidgetFloatable
        )

    def append_log(self, message, level=logging.WARNING):
        QtCore.QMetaObject.invokeMethod(
            self.log_tab,
            "append_log",
            QtCore.Qt.QueuedConnection,
            QtCore.Q_ARG(str, message),
            QtCore.Q_ARG(int, level),
        )


class LogTab(QtWidgets.QWidget):
    """
    tab 'log' in the main widget
    """

    def __init__(self, parent, main_window):
        QtWidgets.QWidget.__init__(self, parent)
        self.main_window = main_window
        self.init_gui()

    def init_gui(self):
        self.vbox = QtWidgets.QVBoxLayout()

        self.log_area = QtWidgets.QTextEdit()
        self.log_area.setReadOnly(True)
        self.log_area.document().setDefaultStyleSheet(STYLE_SHEET)

        self.vbox.addWidget(self.log_area)
        self.setLayout(self.vbox)

    @QtCore.Slot(str, int)
    def append_log(self, message, level=logging.WARNING):
        level_to_style_class = {
            logging.INFO: "info",
            logging.WARNING: "warning",
            logging.ERROR: "error",
        }

        current_date_as_string = datetime.datetime.now().strftime("%m/%d/%y, %H:%M")

        self.log_area.insertHtml(
            '<div><span class="{style_class}">{message_type} <span class="default">({date})</span>:</span> <span>{message}</span></div><br />'.format(
                style_class=level_to_style_class[level],
                message_type=level_to_style_class[level].upper(),
                date=current_date_as_string,
                message=message,
            )
        )

        current_cursor = self.log_area.textCursor()
        current_cursor.movePosition(QtGui.QTextCursor.End)
        self.log_area.setTextCursor(current_cursor)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import sys
from PySide6 import QtCore
from .util.labeled_combobox import LabeledComboBox
from ..config.cutoff_presets import cutoff_presets


class CutoffPresetComboBox(LabeledComboBox):
    preset_selected = QtCore.Signal(object)

    def __init__(self, parent=None):
        super().__init__(label="Preset", parent=parent)
        self._presets = cutoff_presets.presets
        CutoffPresetComboBox._init_ui(self)  # Emulate static binding

    def _init_ui(self):
        for preset in self._presets:
            self.insertItem(len(self._presets), preset.name)
        self.activated.connect(self._on_entry_activated)
        if len(cutoff_presets.presets) == 0:
            self.setVisible(False)

    def discard_preset_choice(self):
        self.setCurrentIndex(0)

    def _on_entry_activated(self, index):
        self.preset_selected.emit(self._presets[index - 1])

    @property
    def selected_preset(self):
        selected_index = self.currentIndex()
        return self._presets[selected_index - 1]



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from PySide6 import QtCore, QtWidgets
import gr
from qtgr import GRWidget
import csv

from ..util.logger import Logger
import sys
import numpy as np
import os
from math import floor


logger = Logger("gui.histogram_widget")
logger.setstream("default", sys.stdout, Logger.DEBUG)


class GrHistogramWidget(GRWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.xvalues = None
        self.yvalues = None
        self.title = None
        self.datapoints = None

    def setdata(self, xvalues, yvalues, widths, title):
        self.xvalues = xvalues
        self.yvalues = yvalues
        self.widths = widths
        self.title = title

    def draw(self, wsviewport=None):
        if self.xvalues is not None and self.widths is not None:
            maxidx = np.argmax(self.xvalues)
            rangex = (self.xvalues.min(), self.xvalues[maxidx] + self.widths[maxidx])
        else:
            rangex = (0.0, 100.0)
        if self.yvalues is not None:
            rangey = gr.adjustrange(0.0, self.yvalues.max())
        else:
            rangey = (0.0, 8.0)

        if wsviewport is None:
            gr.setwsviewport(0, self.mwidth, 0, self.mheight)
        else:
            gr.setwsviewport(*wsviewport)
        gr.setwswindow(0, self.sizex, 0, self.sizey)
        gr.setviewport(
            0.075 * self.sizex, 0.95 * self.sizex, 0.075 * self.sizey, 0.95 * self.sizey
        )
        gr.setwindow(rangex[0], rangex[1], rangey[0], rangey[1])
        gr.setcharheight(0.012)

        gr.setfillintstyle(1)
        gr.setfillcolorind(0)
        gr.fillrect(rangex[0], rangex[1], rangey[0], rangey[1])

        if (
            self.xvalues is not None
            and self.yvalues is not None
            and self.widths is not None
        ):
            gr.setfillintstyle(1)
            gr.setfillcolorind(2)
            for i in range(self.xvalues.size):
                gr.fillrect(
                    self.xvalues[i],
                    self.xvalues[i] + self.widths[i] * 0.8,
                    0.0,
                    self.yvalues[i],
                )
        else:
            gr.text(0.45 * self.sizex, 0.5 * self.sizey, "no data")

        gr.setlinecolorind(1)
        xtick = floor(0.02 * (rangex[1] - rangey[0]) * 100.0) / 100.0
        ytick = floor(0.04 * (rangey[1] - rangey[0]) * 50.0) / 50.0
        gr.axes(xtick, ytick, rangex[0], rangey[0], 10, 5, 0.0075)
        gr.axes(xtick, ytick, rangex[1], rangey[1], -10, -5, -0.0075)

        if self.title is not None:
            gr.text(0.8 * self.sizex, 0.9 * self.sizey, self.title)


class HistogramWidget(QtWidgets.QWidget):

    def __init__(self, parent):
        super().__init__(parent)

        self.control = parent.control
        self.results = None

        self.init_gui()

    def init_gui(self):
        vbox = QtWidgets.QVBoxLayout()
        grid = QtWidgets.QGridLayout()
        self.gr_widget = GrHistogramWidget()

        self.datasetlabel = QtWidgets.QLabel("No data loaded.", self)
        self.datasetlabel.setAlignment(QtCore.Qt.AlignHCenter)

        selectbox = QtWidgets.QHBoxLayout()
        self.cavity_type_box = QtWidgets.QComboBox(self)
        self.cavity_type_box.setMinimumWidth(180)
        selectbox.addWidget(self.cavity_type_box)
        selectbuttongroup = QtWidgets.QButtonGroup(self)
        self.volumebutton = QtWidgets.QRadioButton("Cavity Volume", self)
        selectbox.addWidget(self.volumebutton)
        selectbuttongroup.addButton(self.volumebutton)
        self.areabutton = QtWidgets.QRadioButton("Surface Area", self)
        selectbox.addWidget(self.areabutton)
        selectbuttongroup.addButton(self.areabutton)
        self.volumebutton.setChecked(True)
        grid.addLayout(selectbox, 0, 0)

        self.weightbutton = QtWidgets.QCheckBox("Weighted Histogram", self)
        self.weightbutton.setChecked(True)
        grid.addWidget(self.weightbutton, 0, 1)

        binbox = QtWidgets.QHBoxLayout()
        binbox.addWidget(QtWidgets.QLabel("Number of Bins:", self), 0)
        self.nbins = QtWidgets.QLineEdit(self)
        self.nbins.setMinimumWidth(50)
        binbox.addWidget(self.nbins, 0, QtCore.Qt.AlignLeft)
        grid.addLayout(binbox, 0, 2)

        buttonbox = QtWidgets.QHBoxLayout()

        self.plotbutton = QtWidgets.QPushButton("Plot", self)
        buttonbox.addWidget(self.plotbutton)
        self.plotbutton.clicked.connect(self.draw)

        self.export_image_button = QtWidgets.QPushButton("Save Image", self)
        buttonbox.addWidget(self.export_image_button)
        self.export_image_button.clicked.connect(self.export_image)

        self.export_data_button = QtWidgets.QPushButton("Export Data", self)
        buttonbox.addWidget(self.export_data_button)
        self.export_data_button.clicked.connect(self.export_data)
        grid.addLayout(buttonbox, 1, 0, 1, 3)

        vbox.addWidget(self.gr_widget, stretch=1)
        vbox.addWidget(self.datasetlabel, stretch=0)
        vbox.addLayout(grid)
        self.setLayout(vbox)
        self.show()

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Return or event.key() == QtCore.Qt.Key_Enter:
            self.update()

    def draw(self, now=False, wsviewport=None):
        xvalues = None
        yvalues = None
        widths = None
        title = None
        if self.results is None:
            self.refresh()
        if self.cavity_type_box.currentText():
            if self.cavity_type_box.currentText() == "Surface-based Cavities":
                cavities = self.results.surface_cavities
            elif self.cavity_type_box.currentText() == "Center-based Cavities":
                cavities = self.results.center_cavities
            elif self.cavity_type_box.currentText() == "Cavity Domains":
                cavities = self.results.domains
            nbins = str(self.nbins.text())
            if len(nbins) > 0 and int(nbins) > 0:
                nbins = int(nbins)
            else:
                nbins = 50
            if self.volumebutton.isChecked():
                data = cavities.volumes
            else:
                data = cavities.surface_areas
            if len(data) > 0:
                if self.weightbutton.isChecked():
                    hist, bin_edges = np.histogram(data, bins=nbins, weights=data)
                else:
                    hist, bin_edges = np.histogram(data, bins=nbins)
                widths = bin_edges[1:] - bin_edges[:-1]
                xvalues = bin_edges[:-1]
                yvalues = hist

        self.gr_widget.setdata(xvalues, yvalues, widths, title)
        self.gr_widget.update()
        if now:
            self.gr_widget.draw(wsviewport=wsviewport)

    def export_image(self):
        extensions = (
            ".pdf",
            ".png",
            ".bmp",
            ".jpg",
            ".jpeg",
            ".png",
            ".tiff",
            ".fig",
            ".svg",
            ".wmf",
            ".eps",
            ".ps",
        )
        qtext = "*" + " *".join(extensions)
        filepath = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save Image", ".", "Image Files ({})".format(qtext)
        )[0]
        if not filepath:
            return

        if filepath.endswith(".eps") or filepath.endswith(".ps"):
            gr.beginprintext(filepath, "Color", "A4", "Landscape")
            self.draw(now=True, wsviewport=(0, 0.297 * 0.9, 0, 0.21 * 0.95))
        else:
            gr.beginprint(filepath)
            self.draw(now=True)
        gr.endprint()

    def export_data(self):
        qtext = " *.csv"
        filepath = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save Data", ".", "CSV Files ({})".format(qtext)
        )[0]
        if len(filepath) == 0:
            return
        self.update()
        xvalues = self.gr_widget.xvalues
        yvalues = self.gr_widget.yvalues
        if xvalues is None or yvalues is None:
            return
        with open(filepath, "w") as csvfile:
            csvwriter = csv.writer(csvfile)
            for x, y in zip(xvalues, yvalues):
                csvwriter.writerow([x, y])

    def refresh(self):
        self.cavity_type_box.clear()
        results = self.control.results
        if results is not None:
            results = results[-1][-1]
            if results.surface_cavities is not None:
                self.cavity_type_box.addItem("Surface-based Cavities")
            if results.center_cavities is not None:
                self.cavity_type_box.addItem("Center-based Cavities")
            if results.domains is not None:
                self.cavity_type_box.addItem("Cavity Domains")
            if self.results != results:
                self.results = results
                self.gr_widget.setdata(None, None, None, None)
                self.gr_widget.update()
            self.datasetlabel.setText(str(results))
        else:
            self.datasetlabel.setText("")

    def activate(self):
        self.refresh()

    def updatestatus(self):
        self.refresh()



# coding=utf-8
"""

"""

import sys
import math
import numpy as np
import numpy.linalg as la
import collections


def clamping_acos(cos):
    """
    Calculate arccos with its argument clamped to [-1, 1]
    """
    if cos > 1:
        return 0
    if cos < -1:
        return math.pi / 2
    return math.acos(cos)


def get_bonds_with_constant_delta(atoms, delta):
    """
    Two atoms are connected if their distance is less than or equals a given delta.
    """
    bond_target_index_arrays = []
    for index, position in enumerate(atoms.positions):
        distance_vector_array = atoms.positions[index + 1 :] - position
        distance_squared_array = np.sum(np.square(distance_vector_array), axis=1)
        delta_squared = delta**2
        bond_target_indices = (
            (distance_squared_array <= delta_squared).nonzero()[0] + index + 1
        )
        bond_target_index_arrays.append(bond_target_indices)
    return bond_target_index_arrays


def get_bonds_with_radii(atoms, radii_sum_factor):
    """
    Two atoms are connected if their distance is less than or equals the sum of
    their covalent radii times a radii_sum_factor (e.g. 1.15).
    """
    bond_target_index_arrays = []
    for index, position in enumerate(atoms.positions):
        distance_vector_array = atoms.positions[index + 1 :] - position
        distance_squared_array = np.sum(np.square(distance_vector_array), axis=1)
        delta_squared = np.square(
            (atoms.covalence_radii[index + 1 :] + atoms.covalence_radii[index])
            * radii_sum_factor
        )
        bond_target_indices = (
            (distance_squared_array <= delta_squared).nonzero()[0] + index + 1
        )
        bond_target_index_arrays.append(bond_target_indices)
    return bond_target_index_arrays


def get_bonds_symetric_indicies(bond_target_index_arrays):
    """
    inserts all symetric bonds into the of bonds for each atom
    :param bond_target_index_arrays: list with all symetric bond indices
    :return:
    """
    bond_target_index_arrays = [
        bond_target_index_array.tolist()
        for bond_target_index_array in bond_target_index_arrays
    ]
    for index, bond_target_index_array in enumerate(bond_target_index_arrays):
        for bond_target_index in bond_target_index_array:
            if bond_target_index > index:
                bond_target_index_arrays[bond_target_index].append(index)

    bond_target_index_arrays = np.array(bond_target_index_arrays, dtype=object)
    return bond_target_index_arrays


def calculate_bond_angles(atoms, bond_target_index_arrays):
    # Build dict: atom index -> bonded atom indices
    atom_bonds = collections.defaultdict(list)
    for source_index, target_indices in enumerate(bond_target_index_arrays):
        for target_index in target_indices:
            atom_bonds[source_index].append(target_index)
            atom_bonds[target_index].append(source_index)

    # Calculate angles between bonds sharing an atom
    bond_angles = {}
    for shared_atom_index, target_indices in atom_bonds.items():
        shared_atom_position = atoms.positions[shared_atom_index]
        for i, target_index1 in enumerate(target_indices):
            vec1 = atoms.positions[target_index1] - shared_atom_position
            nvec1 = vec1 / np.linalg.norm(vec1)
            for target_index2 in target_indices[i + 1 :]:
                vec2 = atoms.positions[target_index2] - shared_atom_position
                nvec2 = vec2 / np.linalg.norm(vec2)
                bond_angle = clamping_acos(np.dot(nvec1, nvec2))
                bond_angles[
                    (
                        (target_index1, shared_atom_index),
                        (shared_atom_index, target_index2),
                    )
                ] = bond_angle
                bond_angles[
                    (
                        (target_index2, shared_atom_index),
                        (shared_atom_index, target_index1),
                    )
                ] = bond_angle

    # Calculate bond chains of length 3
    bond_chains = []
    for source_index in atom_bonds.keys():
        bond_chains += find_bond_chains(atoms, source_index, atom_bonds)
    bond_chains_without_duplicates = set()
    for bond_chain in bond_chains:
        if tuple(reversed(bond_chain)) not in bond_chains_without_duplicates:
            bond_chains_without_duplicates.add(tuple(bond_chain))
    bond_chains = bond_chains_without_duplicates

    # Calculate angle in these chains around the axis of the connecting bond
    # according to http://en.wikipedia.org/wiki/Dihedral_angle
    bond_chain_angles = {}
    for bond_chain in bond_chains:
        index1, index2, index3, index4 = bond_chain
        axis = normalized(atoms.positions[index3] - atoms.positions[index2])

        vec1 = atoms.positions[index1] - atoms.positions[index2]
        vec1 -= np.dot(vec1, axis) * axis
        nvec1 = normalized(vec1)

        vec2 = atoms.positions[index4] - atoms.positions[index3]
        vec2 -= np.dot(vec2, axis) * axis
        nvec2 = normalized(vec2)

        angle = clamping_acos(np.dot(nvec1, nvec2))
        if np.dot(nvec2, np.cross(nvec1, axis)) < 0:
            angle = -angle
        bond_chain_angles[tuple(bond_chain)] = angle

    # print "bond_anglges", bond_angles[0]
    return bond_angles, bond_chain_angles


def normalized(vec):
    """Return the normalized version of a numpy array"""
    return vec / la.norm(vec)


def find_bond_chains(
    atoms, source_index, atom_bonds, length=3, previous_bond_chain=None
):
    if previous_bond_chain is None:
        previous_bond_chain = [source_index]
    length -= 1
    bond_chains = []
    for target_index in atom_bonds[source_index]:
        if target_index not in previous_bond_chain:
            new_bond_chain = previous_bond_chain + [target_index]
            new_source_index = target_index
            if length > 0:
                new_bond_chains = find_bond_chains(
                    atoms, new_source_index, atom_bonds, length, new_bond_chain
                )
                if new_bond_chains:
                    bond_chains += new_bond_chains
            else:
                bond_chains.append(new_bond_chain)
    return bond_chains


def export_bonds(filename, atoms):
    bond_target_index_arrays = get_bonds_with_constant_delta(atoms, 2.8)

    with open(filename, "w") as outfile:
        for source_index, target_indices in enumerate(bond_target_index_arrays):
            for target_index in target_indices:
                outfile.write("{} {}\n".format(source_index, target_index))


def export_bond_angles(filename, atoms):
    bond_target_index_arrays = get_bonds_with_constant_delta(atoms, 2.8)
    bond_angles, bond_chain_angles = calculate_bond_angles(
        atoms, bond_target_index_arrays
    )

    with open(filename, "w") as outfile:
        for bond1, bond2 in bond_angles.keys():
            if bond1[0] > bond2[1]:
                outfile.write(
                    "{} {} {} {}\n".format(
                        bond1[0], bond1[1], bond2[1], bond_angles[bond1, bond2]
                    )
                )


def export_bond_dihedral_angles(filename, atoms):
    bond_target_index_arrays = get_bonds_with_constant_delta(atoms, 2.8)
    bond_angles, bond_chain_angles = calculate_bond_angles(
        atoms, bond_target_index_arrays
    )

    with open(filename, "w") as outfile:
        for bond_chain, angle in bond_chain_angles.items():
            outfile.write("{} {} {} {}".format(*bond_chain))
            outfile.write(" {}\n".format(angle))


def main():
    file_name = sys.argv[1]
    frame = int(sys.argv[2])
    f = file.File.open(file_name)
    atoms = f.getatoms(frame)

    bond_target_index_arrays = get_bonds_with_constant_delta(get_atoms(), 2.8)
    bond_target_index_arrays = get_bonds_with_radii(atoms, 1.15)
    bond_angles, bond_chain_angles = calculate_bond_angles(
        get_atoms(), bond_target_index_arrays
    )

    export_bonds("bonds.txt", get_bonds_with_constant_delta(get_atoms(), 2.8))
    export_bond_angles("bond_angles.txt", get_bond_angles())
    export_bond_dihedral_angles("bond_dihedral_angles.txt", get_bond_chain_angles())

    with open("bonds.txt", "w") as outfile:
        for source_index, target_indices in enumerate(bond_target_index_arrays):
            for target_index in target_indices:
                outfile.write("{} {}\n".format(source_index, target_index))
    with open("bond_angles.txt", "w") as outfile:
        for bond1, bond2 in bond_angles.keys():
            if bond1[0] > bond2[1]:
                outfile.write(
                    "{} {} {} {}\n".format(
                        bond1[0], bond1[1], bond2[1], bond_angles[bond1, bond2]
                    )
                )
    with open("bond_dihedral_angles.txt", "w") as outfile:
        for bond_chain, angle in bond_chain_angles.items():
            outfile.write("{} {} {} {}".format(*bond_chain))
            outfile.write(" {}\n".format(angle))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-


from __future__ import absolute_import


import collections
from PySide6 import QtCore, QtGui, QtWidgets
from .table_fit import TableFit


class TableWithRemoveableEntries(QtWidgets.QTableWidget):
    class EntryActionWidget(QtWidgets.QWidget):

        remove = QtCore.Signal()

        def __init__(self, parent=None):
            super().__init__(parent)
            self._init_ui()

        def _init_ui(self):
            self.bt_remove = QtWidgets.QPushButton(
                self.style().standardIcon(QtWidgets.QStyle.SP_TitleBarCloseButton), ""
            )
            self.bt_remove.setFixedSize(30, 30)
            self.bt_remove.setFlat(True)
            self.bt_remove.clicked.connect(lambda event: self.remove.emit())
            self.la_main = QtWidgets.QHBoxLayout()
            self.la_main.addWidget(self.bt_remove)
            self.la_main.setContentsMargins(15, 0, 15, 0)
            self.setLayout(self.la_main)

    def __init__(self, header_labels, entry_count):
        super().__init__(entry_count, len(header_labels) + 1)
        self._table_fit = TableFit(
            self, size_hint_only=True, scrollbar_extra_space=(-1, 2)
        )
        self._header_labels = header_labels
        self._row_index_to_entry_index = range(entry_count)
        self._entry_index_to_row_index = range(entry_count)
        self._indices_to_remove = set()
        TableWithRemoveableEntries._init_ui(self)  # Emulate static binding

    def _init_ui(self):
        self.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Expanding)
        self.setMaximumHeight(300)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalHeaderLabels(self._header_labels + ("",))
        self.setVerticalHeaderLabels(map(str, range(self.entry_count())))
        self.setShowGrid(True)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)

        for i, entry in enumerate(self.entries()):
            for j, item in enumerate(entry):
                if isinstance(item, QtWidgets.QTableWidgetItem):
                    self.setItem(i, j, item)
                else:
                    # item is an arbitrary widget that should be set as cell widget
                    self.setCellWidget(i, j, item)
            entry_action_widget = TableWithRemoveableEntries.EntryActionWidget()
            self.setCellWidget(i, len(entry), entry_action_widget)
            entry_action_widget.remove.connect(
                lambda entry_index=i: self._remove_clicked(entry_index)
            )
        self.resizeRowsToContents()
        self.resizeColumnsToContents()

    def remove_entry(self, row):
        return self.removeRow(row)

    def clear_entries(self):
        self.setRowCount(0)
        self._indices_to_remove.update(range(self.entry_count()))
        self._row_index_to_entry_index = []
        self._entry_index_to_row_index = len(self._entry_index_to_row_index) * [None]

    def removeRow(self, row):
        super(TableWithRemoveableEntries, self).removeRow(row)
        self._indices_to_remove.add(self._row_index_to_entry_index[row])
        del self._row_index_to_entry_index[row]
        self._entry_index_to_row_index = (
            self._entry_index_to_row_index[:row]
            + [None]
            + [
                index - 1 if index is not None else None
                for index in self._entry_index_to_row_index[row + 1 :]
            ]
        )

    def _remove_clicked(self, entry_index):
        self.removeRow(self._entry_index_to_row_index[entry_index])

    def entry_count(
        self,
    ):  # Overriding properties in in a QObject class seems not to work...
        raise NotImplementedError

    def entries(self):
        """Returns a list of entries. A entry is a tuple of QTableWidgetItems (representing a table row).
        Instead of QTableWidgetItems, arbitrary widgets can be used which are set as cell widgets.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""
This module allows the analysis of surface-based molecular cavities in various
volumes. To do so, a volume and a list of atom positions and their cutoff radii
is required. The following example shows reading the first frame of a ``xyz``
file (using the ``pybel`` module), defining a hexagonal volume and moving all
atoms into this volume.

.. code-block:: python
   :emphasize-lines: 7,10

   import pybel

   atoms = pybel.readfile("xyz", "hexagonal.xyz").next().atoms
   num_atoms = len(atoms)
   atom_positions = [atom.coords for atom in atoms]

   volume = volumes.HexagonalVolume(17.68943, 22.61158)
   for atom_index in range(num_atoms):
       atom_positions[atom_index] = volume.get_equivalent_point(atom_positions[atom_index])
   atoms = Atoms(atom_positions, [2.8]*num_atoms)

After this, a discretization of the volume is needed. This module supports
caching of these with the ``DiscretizationCache`` class.

.. code-block:: python

   discretization_cache = DiscretizationCache('cache.hdf5')
   discretization = discretization_cache.get_discretization(volume, 192)

Using this ``discretization`` and the ``Atoms`` object created above, the atom
positions and their cutoff radii are also discretized.

.. code-block:: python

   atom_discretization = AtomDiscretization(atoms, discretization)

With these objects created, all preparations are complete and the domain and
cavity calculation can be done:

.. code-block:: python

   domain_calculation = DomainCalculation(discretization, atom_discretization)
   cavity_calculation = CavityCalculation(domain_calculation)

Additionally, calculation of center-based cavities is possible by passing an
additional parameter to the CavityCalculation:

.. code-block:: python

   cavity_calculation = CavityCalculation(domain_calculation,
                                          use_surface_points=False)

The FakeDomainCalculation class provides a drop-in replacement for the
domain_calculation object in case the results of a previous calculation need to
be used (this is possible as only those attributes which are stored are
actually used during center-based cavity calculation, which is not the case for
surface-based cavity calculations, which at least require the surface point
lists).

The CalculationResults class provides a container for the results and allows
storage to and retrieval from HDF5 files. These files have several groups which
contain the relevant information.

Author: Florian Rhiem <f.rhiem@fz-juelich.de>
"""
from math import pi as PI
import sys
import numpy as np

from ...util import message
from ...util.logger import Logger
from ...computation.split_and_merge.pipeline import start_split_and_merge_pipeline
from ...computation.split_and_merge.algorithm import ObjectType
from ..calculation.gyrationtensor import calculate_gyration_tensor_parameters
from ...util.message import print_message
from .extension import (
    atomstogrid,
    mark_cavities,
    cavity_triangles,
    cavity_intersections,
)


dimension = 3
dimensions = range(dimension)

logger = Logger("core.calculation.algorithm")
logger.setstream("default", sys.stdout, Logger.WARNING)


class DomainCalculation:
    """
    Cavity domain calulation is performed by the following steps:
     1. A grid is created with the resolution defined in the volume
        discretization and filled with zeros.
     2. For each atom, all points in the grid closer to this atom than the
        discrete cavity cutoff radius are set to a point indicating the atom
        index (atom_index+1).
     3. At this point, every point in the grid which is inside of the volume
        and still has a value of zero is part of a cavity domain. To find these
        domains, an optimized split and merge algorithm is applied to the whole
        grid. It returns the center and surface points of each cavity domain
        (points with a neighbor outside of the cavity domain) stored in lists.
        Points inside of a domain are marked with a negative value indicating
        which domain they are part of.
    """

    def __init__(self, discretization, atom_discretization):
        # step 1
        self.discretization = discretization
        self.atom_discretization = atom_discretization
        self.grid = np.zeros(self.discretization.d, dtype=np.int64)

        # step 2
        atomstogrid(
            self.grid,
            self.atom_discretization.discrete_positions,
            self.atom_discretization.atoms.radii_as_indices,
            self.atom_discretization.sorted_discrete_radii,
            [(0, 0, 0)] + self.discretization.combined_translation_vectors,
            self.discretization.grid,
        )
        # step 3
        result = start_split_and_merge_pipeline(
            self.grid,
            self.discretization.grid,
            self.atom_discretization.discrete_positions,
            self.discretization.combined_translation_vectors,
            self.discretization.get_translation_vector,
            ObjectType.DOMAIN,
        )
        (
            self.centers,
            translated_areas,
            non_translated_areas,
            self.surface_point_list,
            self.cyclic_area_indices,
        ) = result
        print_message("Number of domains:", len(self.centers))

        self.domain_volumes = []
        self.critical_domains = (
            []
        )  # count of very small domains -> domains that can disappear on cutoff radius changes
        for domain_index in range(len(self.centers)):
            current_cell_sum = (self.grid == -(domain_index + 1)).sum()
            if current_cell_sum == 1:
                self.critical_domains.append(domain_index)
            domain_volume = current_cell_sum * (self.discretization.s_step**3)
            self.domain_volumes.append(domain_volume)

        self.characteristic_radii = [
            (0.75 * volume / PI) ** (1.0 / 3.0) for volume in self.domain_volumes
        ]

        if translated_areas:
            gyration_tensor_parameters = tuple(
                calculate_gyration_tensor_parameters(area) for area in translated_areas
            )
            (
                self.mass_centers,
                self.squared_gyration_radii,
                self.asphericities,
                self.acylindricities,
                self.anisotropies,
            ) = zip(*gyration_tensor_parameters)
            self.mass_centers = [
                self.discretization.discrete_to_continuous(
                    point, result_inside_volume=True
                )
                for point in self.mass_centers
            ]
            self.squared_gyration_radii = [
                self.discretization.discrete_to_continuous(value, unit_exponent=2)
                for value in self.squared_gyration_radii
            ]
        else:
            (
                self.mass_centers,
                self.squared_gyration_radii,
                self.asphericities,
                self.acylindricities,
                self.anisotropies,
            ) = 5 * ([],)

        self.triangles()

    def triangles(self):
        if hasattr(self, "domain_triangles"):
            return self.domain_triangles
        number_of_domains = len(self.centers)
        print_message("Number of domains:", number_of_domains)
        triangles = []
        surface_areas = []
        step = (self.discretization.s_step,) * 3
        offset = self.discretization.discrete_to_continuous((0, 0, 0))
        for domain_index in range(number_of_domains):
            print_message("Calculating triangles for domain", domain_index)
            vertices, normals, surface_area = cavity_triangles(
                self.grid, [domain_index], 1, step, offset, self.discretization.grid
            )
            triangles.append((vertices, normals))
            surface_areas.append(surface_area)

        self.domain_triangles = triangles
        self.domain_surface_areas = surface_areas
        return triangles


class CavityCalculation:
    """
    Cavity domain calulation is performed by the following steps:
     1. The discrete volume grid is divided into subgrid cells with a side
        length based on the maximum discrete cavity cutoff radius. For each
        subgrid, a tuple of three lists is stored (in self.sg).
     2. The first list for each subgrid cell is filled with the atoms inside
        the cell (their 'real' positions, which might be outside of the
        volume).
     3. The second and third lists are filled with surface points and their
        domain index. (These might also be moved with the translation vectors
        and might thereby also be outside of the volume.)
     4. A new grid is created (grid3) and each point in this grid is set to
        zero if it is outside of the volume or part inside of the cavity cutoff
        radius of an atom, or a negative value if it is part of a cavity domain
        (see the domain calculation step 2 for this).
     5. For each point which is inside the cavity cutoff radius of an
        atom, the nearest atom and the nearest domain surface point are found
        by using the neighbor subgrid cells. If a domain surface point is
        nearer than the nearest atom, than the point belongs to the cavity
        which this surface point belonged to and is marked with a negative
        value.
     6. At this point, two cavities constructed from two cavity domains might
        actually be one multicavity. In this step, these are found and a list
        of multicavities is created.

    About the subgrid cells:
    If a point inside a subgrid cell was marked as 'near an atom' during the
    domain calculation, then the position of this atom must be either in the
    same cell or in one of the cell's neighbors. This is guaranteed, because
    the atom must be at most its on cavity cutoff radius away, and the subgrid
    cell size is the maximum cavity cutoff radius.


    To calculate center-based cavities, use a grid filled with zeros instead of
    resuing some values from the domain calculation grid and then iterate over
    the domain centers instead of the domain surface points.
    """

    def __init__(
        self,
        domain_calculation,
        use_surface_points=True,
        gyration_tensor_parameters=False,
    ):
        self.domain_calculation = domain_calculation
        if use_surface_points:
            self.grid = self.domain_calculation.grid
            num_surface_points = sum(
                map(len, self.domain_calculation.surface_point_list)
            )
            print_message("Number of surface points:", num_surface_points)
        else:
            self.grid = None

        self.sg_cube_size = (
            self.domain_calculation.atom_discretization.sorted_discrete_radii[0]
        )
        if use_surface_points:
            domain_seed_point_lists = self.domain_calculation.surface_point_list
        else:
            domain_seed_point_lists = [
                [center] for center in self.domain_calculation.centers
            ]

        discretization = self.domain_calculation.discretization
        atom_discretization = self.domain_calculation.atom_discretization

        # steps 1 to 5
        self.grid3 = mark_cavities(
            self.grid,
            discretization.grid,
            discretization.d,
            self.sg_cube_size,
            atom_discretization.discrete_positions,
            [(0, 0, 0)] + discretization.combined_translation_vectors,
            domain_seed_point_lists,
            use_surface_points,
        )

        if gyration_tensor_parameters:
            result = start_split_and_merge_pipeline(
                self.grid3,
                discretization.grid,
                atom_discretization.discrete_positions,
                discretization.combined_translation_vectors,
                discretization.get_translation_vector,
                ObjectType.CAVITY,
            )
            translated_areas, non_translated_areas, cyclic_area_indices = result

        num_domains = len(self.domain_calculation.centers)
        grid_volume = (discretization.grid == 0).sum()
        self.cavity_volumes = []
        for domain_index in range(num_domains):
            self.cavity_volumes.append(
                1.0
                * (self.grid3 == -(domain_index + 1)).sum()
                * (discretization.s_step**3)
            )
        if len(self.cavity_volumes) > 0:
            volume = self.cavity_volumes[0]
        self.characteristic_radii = [
            (0.75 * volume / PI) ** (1.0 / 3.0) for volume in self.cavity_volumes
        ]

        # step 6
        intersection_table = cavity_intersections(self.grid3, num_domains)
        multicavities = []
        cavity_to_neighbors = num_domains * [None]
        for domain in range(num_domains):
            current_neighbors = set([domain])
            for neighbor in range(num_domains):
                if intersection_table[domain][neighbor] == 1:
                    current_neighbors.add(neighbor)
            for multicavity in multicavities[:]:
                if any([neighbor in multicavity for neighbor in current_neighbors]):
                    current_neighbors = current_neighbors | multicavity
                    multicavities.remove(multicavity)
            multicavities.append(current_neighbors)
            for neighbor in current_neighbors:
                cavity_to_neighbors[neighbor] = current_neighbors
        self.multicavities = multicavities
        self.multicavity_volumes = []
        for multicavity in multicavities:
            self.multicavity_volumes.append(
                sum(self.cavity_volumes[cavity_index] for cavity_index in multicavity)
            )
        print_message("Multicavity volumes:", self.multicavity_volumes)

        if gyration_tensor_parameters:
            if len(self.multicavities) == len(
                translated_areas
            ):  # TODO check weather split and merge and multicavity intersection give the same result for multicavities
                # `self.multicavities` entries are sorted by the largest contained neighbor index. Thus sort the indices to
                # access `non_translated_areas` and `translated_areas` to match the order of `self.multicavities`.
                def key_func(cavity_index):
                    cavity_area = non_translated_areas[cavity_index]
                    a_single_cavity_index = -self.grid3[cavity_area[0]] - 1
                    max_neighbor_index = max(cavity_to_neighbors[a_single_cavity_index])
                    return max_neighbor_index

                sorted_area_indices = sorted(
                    range(len(self.multicavities)), key=key_func
                )
                sorted_translated_areas = [
                    translated_areas[i] for i in sorted_area_indices
                ]
                sorted_cyclic_area_indices = [
                    i
                    for i, index in enumerate(sorted_area_indices)
                    if index in cyclic_area_indices
                ]
                self.cyclic_area_indices = sorted_cyclic_area_indices

                gyration_tensor_parameters = tuple(
                    calculate_gyration_tensor_parameters(area)
                    for area in sorted_translated_areas
                )
                (
                    self.mass_centers,
                    self.squared_gyration_radii,
                    self.asphericities,
                    self.acylindricities,
                    self.anisotropies,
                ) = zip(*gyration_tensor_parameters)
                self.mass_centers = [
                    discretization.discrete_to_continuous(
                        point, result_inside_volume=True
                    )
                    for point in self.mass_centers
                ]
                self.squared_gyration_radii = [
                    discretization.discrete_to_continuous(value, unit_exponent=2)
                    for value in self.squared_gyration_radii
                ]
            else:
                logger.warn(
                    "Gyration tensors could not be calculated try increasing the resolution."
                )
                message.log(
                    "Gyration tensors could not be calculated try increasing the resolution."
                )
                (
                    self.mass_centers,
                    self.squared_gyration_radii,
                    self.asphericities,
                    self.acylindricities,
                    self.anisotropies,
                ) = 5 * ([],)

        self.triangles()

    def squared_distance(self, a, b):
        """
        Calculates the squared distance between two points while taking the
        translation vectors into account.
        """
        sqd = sys.maxsize
        for v in self.domain_calculation.discretization.combined_translation_vectors + [
            (0, 0, 0)
        ]:
            sqd = min(
                sqd,
                sum([(a[i] - b[i] + v[i]) * (a[i] - b[i] + v[i]) for i in dimensions]),
            )
        return sqd

    def triangles(self):
        if hasattr(self, "cavity_triangles"):
            return self.cavity_triangles
        step = (self.domain_calculation.discretization.s_step,) * 3
        offset = self.domain_calculation.discretization.discrete_to_continuous(
            (0, 0, 0)
        )
        triangles = []
        surface_areas = []
        for i, multicavity in enumerate(self.multicavities):
            print_message("Generating triangles for multicavity:", i + 1)
            vertices, normals, surface_area = cavity_triangles(
                self.grid3,
                multicavity,
                4,
                step,
                offset,
                self.domain_calculation.discretization.grid,
            )
            triangles.append((vertices, normals))
            surface_areas.append(surface_area)

        self.cavity_triangles = triangles
        self.cavity_surface_areas = surface_areas
        return cavity_triangles

    def __getattr__(self, attr):
        optional_attributes = (
            "mass_centers",
            "squared_gyration_radii",
            "asphericities",
            "acylindricities",
            "anisotropies",
            "characteristic_radii",
            "cyclic_area_indices",
        )
        if attr in optional_attributes:
            return None
        else:
            return super(CavityCalculation, self).__getattr__(attr)


class FakeDomainCalculation(object):
    """
    When calculating center-based cavities, a DomainCalculation object is
    required. This has to provide the domain central points, but not the surface
    points (as those are only needed for surface-based cavity calculation).
    Objects of this class can be used as a drop-in for 'real'
    DomainCalculations, e.g. when the required data is loaded from a file.
    """

    def __init__(self, discretization, atom_discretization, results):
        self.centers = results.domains.centers
        self.discretization = discretization
        self.atom_discretization = atom_discretization



# -*- coding: utf-8 -*-
"""
Visualize Atoms and Cavities with GR3
"""

from math import sin, cos, sqrt, pi
import gr3
from ..config.configuration import config
from ..core import calculation as calculation
import numpy as np
import numpy.linalg as la
import os
from ctypes import c_int
from ..util.gl_util import (
    create_perspective_projection_matrix,
    create_look_at_matrix,
    create_rotation_matrix_homogenous,
    create_translation_matrix_homogenous,
)


class Visualization(object):
    """
    Visualize Atoms and Cavities with GR3
    """

    def __init__(self):
        self.max_side_lengths = 1.0
        self.mat = np.eye(4)
        self.d = self.max_side_lengths * 2
        self.pos = np.array((0, 0, self.d))
        self.up = np.array((0, 1, 0))
        self.right = np.array((1, 0, 0))

        self.near = 0.1
        self.far = 3 * self.d
        self.fov = 45.0

        self.width = 0
        self.height = 0
        self.usecurrentframebuffer = False
        #  self.assigned_opengl_context = None

        self.results = None
        self.settings = VisualizationSettings()
        self.objectids = [None]

    def setresults(self, results):
        """
        Set the results that will be displayed

        **Parameters:**
            `results` :
                :class:`core.data.Results` object to visualize
        """
        self.results = results
        max_side_lengths = max(results.atoms.volume.side_lengths)
        self.d = self.d / self.max_side_lengths * max_side_lengths
        self.max_side_lengths = max_side_lengths
        self.far = 6 * max_side_lengths
        self.create_scene()
        if self.width != 0 and self.height != 0:
            self.paint(self.width, self.height)
        self.settings.visible_domain_indices = None
        self.settings.visible_surface_cavity_indices = None
        self.settings.visible_center_cavity_indices = None

    def create_scene(self):
        """
        Create GR3 meshes. ``self.results`` contains the mesh data
        and in ``self.settings`` is specified which meshes shound be rendered.
        """

        c = config.Colors.background
        gr3.setbackgroundcolor(c[0], c[1], c[2], 1.0)
        gr3.clear()

        if self.results is None:
            return

        show_domains = self.settings.show_domains
        show_surface_cavities = self.settings.show_surface_cavities
        show_center_cavities = self.settings.show_center_cavities
        if show_center_cavities and self.results.center_cavities is not None:
            show_surface_cavities = False
        elif show_surface_cavities and self.results.surface_cavities is not None:
            show_domains = False

        self.objectids = [None]
        edges = self.results.atoms.volume.edges
        num_edges = len(edges)
        edge_positions = [edge[0] for edge in edges]
        edge_directions = [
            [edge[1][i] - edge[0][i] for i in range(3)] for edge in edges
        ]
        edge_lengths = [sum([c * c for c in edge]) ** 0.5 for edge in edge_directions]
        edge_radius = min(edge_lengths) / 200
        if self.settings.show_bounding_box:
            gr3.drawcylindermesh(
                num_edges,
                edge_positions,
                edge_directions,
                [config.Colors.bounding_box] * num_edges,
                [edge_radius] * num_edges,
                edge_lengths,
            )
            corners = list(
                set(
                    [tuple(edge[0]) for edge in edges]
                    + [tuple(edge[1]) for edge in edges]
                )
            )
            num_corners = len(corners)
            gr3.drawspheremesh(
                num_corners,
                corners,
                [config.Colors.bounding_box] * num_corners,
                [edge_radius] * num_corners,
            )

        if self.settings.show_atoms and self.results.atoms is not None:
            visible_atom_indices = self.settings.visible_atom_indices
            if visible_atom_indices is not None:
                visible_atom_indices = [
                    comp
                    for comp in visible_atom_indices
                    if 0 <= comp < self.results.atoms.number
                ]
            else:
                visible_atom_indices = range(self.results.atoms.number)
            if len(visible_atom_indices) == 0:
                visible_atom_indices = None
            if visible_atom_indices is not None:
                visible_atom_indices = np.array(visible_atom_indices)
                gr3.drawspheremesh(
                    len(visible_atom_indices),
                    self.results.atoms.positions[visible_atom_indices],
                    self.results.atoms.colors[visible_atom_indices],
                    np.ones(len(visible_atom_indices)) * config.OpenGL.atom_radius,
                )
                if self.settings.show_bonds:
                    bonds = self.results.atoms.bonds
                    for start_index, target_indices in enumerate(bonds):
                        if start_index not in visible_atom_indices:
                            continue
                        target_indices = np.array(
                            [i for i in target_indices if i in visible_atom_indices]
                        )
                        if len(target_indices) == 0:
                            continue
                        start_position = self.results.atoms.positions[start_index]
                        target_positions = self.results.atoms.positions[target_indices]
                        directions = target_positions - start_position
                        bond_lengths = la.norm(directions, axis=1)
                        directions /= bond_lengths.reshape(len(directions), 1)
                        gr3.drawcylindermesh(
                            len(target_indices),
                            target_positions,
                            -directions,
                            [config.Colors.bonds] * self.results.atoms.number,
                            np.ones(bond_lengths.shape) * config.OpenGL.bond_radius,
                            bond_lengths,
                        )

        if self.results is None:
            return
        if show_domains and self.results.domains is not None:
            self.draw_cavities(
                self.results.domains,
                config.Colors.domain,
                "domain",
                self.settings.visible_domain_indices,
            )
        if show_surface_cavities and self.results.surface_cavities is not None:
            self.draw_cavities(
                self.results.surface_cavities,
                config.Colors.surface_cavity,
                "surface cavity",
                self.settings.visible_surface_cavity_indices,
            )
        if show_center_cavities and self.results.center_cavities is not None:
            self.draw_cavities(
                self.results.center_cavities,
                config.Colors.center_cavity,
                "center cavity",
                self.settings.visible_center_cavity_indices,
            )

    def draw_cavities(self, cavities, color, cavity_type, indices=None):
        if indices is None:
            indices = range(self.results.domains.number)
        for index in set(indices):
            if 0 <= index < len(cavities.triangles):
                triangles = cavities.triangles[index]
                gr3._gr3.gr3_setobjectid(gr3.c_int(len(self.objectids)))
                self.objectids.append((cavity_type, index))
                mesh = gr3.createmesh(
                    triangles.shape[1] * 3,
                    triangles[0, :, :, :],
                    triangles[1, :, :, :],
                    [color] * (triangles.shape[1] * 3),
                )
                gr3.drawmesh(
                    mesh, 1, (0, 0, 0), (0, 0, 1), (0, 1, 0), (1, 1, 1), (1, 1, 1)
                )
                gr3.deletemesh(c_int(mesh.value))
        gr3._gr3.gr3_setobjectid(gr3.c_int(0))

    def zoom(self, delta):
        """
        Adjust the zoom level.

        **Parameters:**
            `delta` :
                the increment of the distance from the camera to the center
        """
        zoom_v = 1.0 / 20
        zoom_cap = self.d + zoom_v * delta < (self.max_side_lengths) * 4
        if self.d + zoom_v * delta > 0 and zoom_cap:
            self.d += zoom_v * delta

    def translate_mouse(self, dx, dy):
        """
        Translate the model.
        """
        trans_v = 1.0 / 1000
        trans_factor = trans_v * max(self.d, 20)
        m = create_translation_matrix_homogenous(
            -dx * trans_factor, dy * trans_factor, 0
        )
        self.mat = self.mat.dot(m)

    def set_focus_on(self, x, y, z):
        self.mat[:3, 3] = (x, y, z)

    def rotate_mouse(self, dx, dy):
        """
        Rotate the model according to a mouse movement (dx, dy) on the screen.
        """
        rot_v = 1.0 / 13000
        if dx == 0 and dy == 0:
            return
        rot_axis = np.array((-dy, -dx, 0.0))
        rot_factor = max(self.d, 20) * rot_v * la.norm(rot_axis)
        # rotation matrix with min rotation angle
        m = create_rotation_matrix_homogenous(
            rot_factor, rot_axis[0], rot_axis[1], rot_axis[2]
        )
        self.mat = self.mat.dot(m)

    def reset_view(self):
        self.mat = np.eye(4)

    def set_camera(self, width, height):
        """
        Update the shown scene after the perspective has changed.
        """
        rightt = self.mat[:3, 0]
        upt = self.mat[:3, 1]
        pt = self.mat[:3, 2] * self.d
        t = self.mat[:3, 3]

        self.proj_mat = create_perspective_projection_matrix(
            np.radians(self.fov), 1.0 * width / height, self.near, self.far
        )
        gr3.setcameraprojectionparameters(self.fov, self.near, self.far)
        self.lookat_mat = create_look_at_matrix(pt + t, t, upt)
        gr3.cameralookat(
            pt[0] + t[0],
            pt[1] + t[1],
            pt[2] + t[2],
            t[0],
            t[1],
            t[2],
            upt[0],
            upt[1],
            upt[2],
        )

    def assign_opengl_context(self, context):
        self.assigned_opengl_context = context

    def paint(self, width, height, usecurrentframebuffer=None, device_pixel_ratio=1):
        """
        Refresh the OpenGL scene.
        """
        self.width = width
        self.height = height
        self.set_camera(width, height)
        if usecurrentframebuffer is not None:
            self.usecurrentframebuffer = usecurrentframebuffer
        if self.usecurrentframebuffer:
            gr3.usecurrentframebuffer()
        gr3.drawimage(
            0,
            width * device_pixel_ratio,
            0,
            height * device_pixel_ratio,
            width,
            height,
            gr3.GR3_Drawable.GR3_DRAWABLE_OPENGL,
        )

    def save_screenshot(
        self, file_name, width=3840, height=2160, first=True, last=True
    ):
        """
        Save a screenshot in the given resolution.
        ``first`` and ``last`` can be used to indicate if the first or last screenshot is taken when
        multiple images are saved in a loop for example.
        """
        gr3.export(file_name, width, height)

    def get_object_at_2dposition(self, x, y):
        oid = gr3.c_int(0)
        gr3._gr3.gr3_selectid(
            gr3.c_int(x),
            gr3.c_int(y),
            gr3.c_int(self.width),
            gr3.c_int(self.height),
            gr3.byref(oid),
        )
        return self.objectids[oid.value]

    def get_object_at_3dposition(self, x, y, z):
        if self.results is None:
            return None
        if self.results.atoms is not None:
            positions = self.results.atoms.positions
            distances = la.norm(positions - (x, y, z), axis=1)
            nearest_atom_index = np.argmin(distances)

            # calculate atom radius
            edges = self.results.atoms.volume.edges
            edge_directions = [
                [edge[1][i] - edge[0][i] for i in range(3)] for edge in edges
            ]
            edge_lengths = [
                sum([c * c for c in edge]) ** 0.5 for edge in edge_directions
            ]
            edge_radius = min(edge_lengths) / 200
            atom_radius = 4 * edge_radius

            if 0.95 < distances[nearest_atom_index] / atom_radius < 1.05:
                return ("atom", nearest_atom_index)


class VisualizationSettings(object):
    """
    Container for settings that control which part of the data is visualized.
    Attributes:

        `show_domains`

        `show_surface_cavities`

        `show_center_cavities`

        `show_atoms`

        `show_bounding_box`

    """

    def __init__(
        self,
        domains=False,
        show_surface_cavities=True,
        show_center_cavities=False,
        atoms=True,
        bonds=True,
        bounding_box=True,
        visible_atom_indices=None,
        visible_domain_indices=None,
        visible_surface_cavity_indices=None,
        visible_center_cavity_indices=None,
    ):
        self.show_domains = domains
        self.visible_domain_indices = visible_domain_indices
        self.show_surface_cavities = show_surface_cavities
        self.visible_surface_cavity_indices = visible_surface_cavity_indices
        self.show_center_cavities = show_center_cavities
        self.visible_center_cavity_indices = visible_center_cavity_indices
        self.show_atoms = atoms
        self.visible_atom_indices = visible_atom_indices
        self.show_bonds = bonds
        self.show_bounding_box = bounding_box
